# Initialise the logger
import logging
from collections import defaultdict

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


MIN_CNT_TO_SHOW_ON_PLOT = 5


class DragenFragmentLength(BaseMultiqcModule):
    def add_fragment_length_hist(self):
        data_by_rg_by_sample = defaultdict(dict)

        for f in self.find_log_files("dragen/fragment_length_hist"):
            data_by_rg = parse_fragment_length_hist_file(f)
            s_name = f["s_name"]
            if s_name in data_by_rg_by_sample:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.add_data_source(f, section="fragment_length_hist")

            for rg, data in data_by_rg.items():
                if any(rg in d_rg for sn, d_rg in data_by_rg_by_sample.items()):
                    log.debug(f"Duplicate read group name {rg} found for {s_name}! Overwriting")
            data_by_rg_by_sample[s_name].update(data_by_rg)

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, s_name)

        # Filter to strip out ignored sample names:
        data_by_rg_by_sample = self.ignore_samples(data_by_rg_by_sample)
        if not data_by_rg_by_sample:
            return set()

        # Write data to file
        self.write_data_file(data_by_rg_by_sample, "dragen_frag_len")

        # Merging all data
        data_by_rg = {}
        for sn, d_rg in data_by_rg_by_sample.items():
            for rg, d in d_rg.items():
                if rg in d_rg:
                    rg = rg + " (" + sn + ")"
                data_by_rg[rg] = d

        # Exit early if we have no valid data, such as from FastQcOnly runs
        if not data_by_rg:
            return set()

        smooth_points = 300
        self.add_section(
            name="Fragment length hist",
            anchor="dragen-fragment-length-histogram",
            description="""
            Distribution of estimated fragment lengths of mapped reads per read group.
            Only points supported by at least {} reads are shown to prevent long flat tail.
            The plot is also smoothed down to showing 300 points on the X axis to reduce noise.
            """.format(MIN_CNT_TO_SHOW_ON_PLOT),
            plot=linegraph.plot(
                data_by_rg,
                {
                    "id": "dragen_fragment_length",
                    "title": "Dragen: Fragment length hist",
                    "ylab": "Number of reads",
                    "xlab": "Fragment length (bp)",
                    "ymin": 0,
                    "xmin": 0,
                    "tt_label": "<b>{point.x} bp</b>: {point.y} reads",
                    "smooth_points": smooth_points,
                },
            ),
        )
        return data_by_rg_by_sample.keys()


def parse_fragment_length_hist_file(f):
    """
    T_SRR7890936_50pc.fragment_length_hist.csv

    #Sample: N_SRR7890889
    FragmentLength,Count
    36,1
    37,0
    38,0
    39,0
    40,0
    41,1
    ...
    39203,0
    39204,0
    39205,1
    #Sample: T_SRR7890936_50pc
    FragmentLength,Count
    53,2
    54,0
    ...
    39316,0
    39317,1
    """

    data_by_rg = defaultdict(dict)

    read_group = None
    for line in f["f"].splitlines():
        if line.startswith("#Sample"):
            read_group = line.split("#Sample: ")[1]
        else:
            assert read_group is not None
            frag_len, cnt = line.split(",")
            try:
                frag_len = int(frag_len)
                cnt = int(cnt)
            except ValueError:
                assert line == "FragmentLength,Count", line
            else:
                if cnt >= MIN_CNT_TO_SHOW_ON_PLOT:  # to prevent long flat tail
                    data_by_rg[read_group][frag_len] = cnt

    return data_by_rg



## Nicolas Servant
## April 2018

""" MultiQC module to parse output from HiC-Pro """


import logging
import os.path

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """HiC-Pro module, parses log and stats files saved by HiC-Pro."""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="HiC-Pro",
            anchor="hicpro",
            href="https://github.com/nservant/HiC-Pro",
            info="is an efficient and flexible pipeline for Hi-C data processing. The MultiQC module is supported since HiC-Pro v2.11.0.",
            doi="10.1186/s13059-015-0831-x",
        )

        # Find and load any HiC-Pro summary reports
        self.hicpro_data = dict()
        for k in ["mmapstat", "mpairstat", "mergestat", "mRSstat", "assplit"]:
            for f in self.find_log_files(f"hicpro/{k}"):
                self.parse_hicpro_stats(f, k)

        # Update current statistics
        for s_name in self.hicpro_data:
            data = self.hicpro_data[s_name]
            try:
                if "valid_interaction" in data:
                    if "valid_interaction_rmdup" in data:
                        data["duplicates"] = data["valid_interaction"] - data["valid_interaction_rmdup"]
                        if float(data["valid_interaction"]) > 0:
                            data["percent_duplicates"] = (
                                float(data["duplicates"]) / float(data["valid_interaction"]) * 100.0
                            )
                    if "valid_pairs_on_target" in data:
                        data["valid_pairs_off_target"] = int(data["valid_interaction"]) - int(
                            data["valid_pairs_on_target"]
                        )
                    if "total_R1" in data and float(data["total_R1"]) > 0:
                        data["percent_valid"] = float(data["valid_interaction"]) / float(data["total_R1"]) * 100.0

                if "Reported_pairs" in data:
                    data["paired_reads"] = int(data["Reported_pairs"])
                    if "total_R1" in data and float(data["total_R1"]) > 0:
                        data["percent_paired_reads"] = float(data["Reported_pairs"]) / float(data["total_R1"]) * 100.0

                if "mapped_R1" in data:
                    data["Failed_To_Align_Read_R1"] = int(data["total_R1"]) - int(data["mapped_R1"])
                    if "total_R1" in data and float(data["total_R1"]) > 0:
                        data["percent_mapped_R1"] = float(data["mapped_R1"]) / float(data["total_R1"]) * 100.0

                if "mapped_R2" in data:
                    data["Failed_To_Align_Read_R2"] = int(data["total_R2"]) - int(data["mapped_R2"])
                    if "total_R2" in data and float(data["total_R2"]) > 0:
                        data["percent_mapped_R2"] = float(data["mapped_R2"]) / float(data["total_R2"]) * 100.0

            except KeyError as e:
                log.error(f"Missing expected key {e} in sample '{s_name}'")
            except (ValueError, TypeError):
                pass

        # Filter to strip out ignored sample names
        self.hicpro_data = self.ignore_samples(self.hicpro_data)

        if len(self.hicpro_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.hicpro_data)} HiC-Pro reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed data to a file
        self.write_data_file(self.hicpro_data, "multiqc_hicpro")

        # Basic Stats Table
        self.hicpro_stats_table()

        # Report sections
        self.add_section(
            name="Read Mapping",
            anchor="hicpro-mapping",
            description="Alignment of reads in single-end mode.",
            helptext="""
            HiC-Pro uses a two steps mapping strategy. End-to-end mapping is first
            performed (*Full reads Alignments*). Unmapped reads are then trimmed
            at their ligation site, and the 5\' end is re-aligned on the reference
            genome (*Trimmed Reads Alignments*)""",
            plot=self.hicpro_mapping_chart(),
        )

        self.add_section(
            name="Read Pairing",
            anchor="hicpro-pairing",
            description="Pairing of single-end mapping.",
            helptext="""
            HiC-Pro combines aligned reads as pairs. Singleton, low quality pairs or
            pairs involving multi-mapped reads are usually discarded. Note that the filtering
            at pairing level can change accrding to the parameters used.""",
            plot=self.hicpro_pairing_chart(),
        )

        self.add_section(
            name="Read Pair Filtering",
            anchor="hicpro-filtering",
            description="Filtering of read pairs.",
            helptext="""
            Aligned read pairs are filtered to select valid 3C products from two
            different restriction fragments. Read pairs coming from the same fragments,
            such as self-ligation or unligated (danging-end) fragments, are discarded.
            Ligation products involving neighboring restriction fragment (religation) are also discarded.
            Filtered pairs correspond to ligation products discarded based on the range of insert/fragment sizes defined
            in the analysis. Finally, as the ligation should be a random process, valid read pairs from all
            orientations (R=Reverse, F=forward) are expected to be observed in the same proportion.""",
            plot=self.hicpro_filtering_chart(),
        )

        self.add_section(
            name="Contact Statistics",
            anchor="hicpro-rmdup",
            description="Contacts statistics after duplicates removal.",
            helptext="""
            Description of contact frequency after duplicates removal. Intra-chromosomal
            (*cis*) interaction are expected to be more frequent than inter-chromosomal contacts (*trans*)""",
            plot=self.hicpro_contact_chart(),
        )

        allele_plot = self.hicpro_as_chart()
        if allele_plot:
            self.add_section(
                name="Allele-specific analysis",
                anchor="hicpro-asan",
                description="Assignment of valid interactions (afer duplicates removal) to parental alleles.",
                helptext="""
                Description of allele-specific contacts. Valid interactions (0-1 and 1-1, resp. 0-2, 2-2)
                are used to build the genome1 (resp. genome2) parental maps.""",
                plot=allele_plot,
            )

        capture_plot = self.hicpro_capture_chart()
        if capture_plot:
            self.add_section(
                name="Capture analysis",
                anchor="hicpro-cap",
                description="Selection of interactions overlapping the targeted region(s).",
                helptext="""
                Description of capture efficiency. Valid interactions with either two (capture-capture) or
                one (capture-reporter) interactors overlapping with the target(s) are reported.""",
                plot=capture_plot,
            )

    def parse_hicpro_stats(self, f, rsection):
        """Parse a HiC-Pro stat file"""
        s_name = self.clean_s_name(os.path.basename(f["root"]), root=os.path.dirname(f["root"]))
        if s_name not in self.hicpro_data.keys():
            self.hicpro_data[s_name] = {}

        self.add_data_source(f, s_name, section=rsection)
        for line in f["f"].splitlines():
            if not line.startswith("#"):
                s = line.split("\t")
                if s[0] in self.hicpro_data[s_name]:
                    log.debug(
                        f"Duplicated key {s[0]} found in {f['root'] + '/' + f['fn']} for sample {s_name}, overwriting"
                    )
                # Try to convert the extracted value to a number and store it in hicpro_data.
                # try-block is used to prevent program crash, because there is no
                # guarantee that the value (s[1]) can be always converted to integer.
                try:
                    self.hicpro_data[s_name][s[0]] = int(s[1])
                except ValueError:
                    # Convert to float (also works for inf and scientific (exponential) notation).
                    try:
                        self.hicpro_data[s_name][s[0]] = float(s[1])
                    # Otherwise just store the value as is.
                    except ValueError:
                        log.error(f"Could not parse value for key '{s[0]}' as a number in sample '{s_name}': '{s[1]}'")
                        self.hicpro_data[s_name][s[0]] = s[1]

    def hicpro_stats_table(self):
        """Add HiC-Pro stats to the general stats table"""
        headers = {
            "percent_duplicates": {
                "title": "% Duplicates",
                "description": "Percent of duplicated valid pairs (%)",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlOrRd",
                "hidden": True,
            },
            "valid_interaction_rmdup": {
                "title": f"{config.read_count_prefix} Valid Pairs Unique",
                "description": f"Number of valid pairs after duplicates removal ({config.read_count_desc})",
                "min": 0,
                "scale": "RdYlBu",
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
            },
            "percent_valid": {
                "title": "% Valid Pairs",
                "description": "Percentage of valid pairs over reported ones (%)",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn",
                "hidden": True,
            },
            "valid_interaction": {
                "title": f"{config.read_count_prefix} Valid Pairs",
                "description": f"Number of valid pairs ({config.read_count_desc})",
                "min": 0,
                "scale": "RdYlBu",
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
                "hidden": True,
            },
            "percent_paired_reads": {
                "title": "% Reported",
                "description": "Percentage of paired reads (%) passing the mapping filters",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn",
                "hidden": True,
            },
            "paired_reads": {
                "title": "Reported Read Pairs",
                "description": "Total number of read pairs ({}) passing the mapping filters".format(
                    config.read_count_desc
                ),
                "min": 0,
                "scale": "RdYlBu",
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
            },
            "percent_mapped_R2": {
                "title": "% Aligned [R2]",
                "description": "Percentage of aligned reads [R2] (%)",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn",
                "hidden": True,
            },
            "mapped_R2": {
                "title": "Aligned [R2]",
                "description": f"Total number of aligned reads [R2] ({config.read_count_desc})",
                "min": 0,
                "scale": "RdYlBu",
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
                "hidden": True,
            },
            "percent_mapped_R1": {
                "title": "% Aligned [R1]",
                "description": "Percentage of aligned reads [R1] (%)",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn",
                "hidden": True,
            },
            "mapped_R1": {
                "title": "Aligned [R1]",
                "description": f"Total number of aligned reads [R1] ({config.read_count_desc})",
                "min": 0,
                "scale": "RdYlBu",
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
                "hidden": True,
            },
            "total_R1": {
                "title": "Total",
                "description": "Total Number of Read Pairs",
                "min": 0,
                "scale": "RdYlBu",
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
            },
        }

        self.general_stats_addcols(self.hicpro_data, headers)

    def hicpro_mapping_chart(self):
        """Generate the HiC-Pro Aligned reads plot"""

        # Specify the order of the different possible categories
        keys = {
            "Full_Alignments_Read": {"color": "#005ce6", "name": "Full reads Alignments"},
            "Trimmed_Alignments_Read": {"color": "#3385ff", "name": "Trimmed reads Alignments"},
            "Failed_To_Align_Read": {"color": "#a9a2a2", "name": "Failed To Align"},
        }

        data = [{}, {}]
        for s_name in self.hicpro_data:
            for r in [1, 2]:
                n_global = self.hicpro_data[s_name].get(f"global_R{r}")
                n_local = self.hicpro_data[s_name].get(f"local_R{r}")
                n_mapped = self.hicpro_data[s_name].get(f"mapped_R{r}")
                n_total = self.hicpro_data[s_name].get(f"total_R{r}")
                d = dict()
                if n_global is not None:
                    d["Full_Alignments_Read"] = n_global
                if n_local is not None:
                    d["Trimmed_Alignments_Read"] = n_local
                if n_mapped is not None and n_total is not None:
                    try:
                        d["Failed_To_Align_Read"] = int(n_total) - int(n_mapped)
                    except ValueError:
                        log.warning(
                            f"Could not parse mapped_R{r}={n_mapped} or total_R{r}={n_total} "
                            f"as an integer number for sample {s_name}"
                        )
                data[r - 1][f"{s_name} [R{r}]"] = d

        # Config for the plot
        config = {
            "id": "hicpro_mapping_stats_plot",
            "title": "HiC-Pro: Mapping Statistics",
            "ylab": "# Reads: Read 1",
            "data_labels": [
                {"name": "Read 1", "ylab": "# Reads: Read 1"},
                {"name": "Read 2", "ylab": "# Reads: Read 2"},
            ],
        }

        return bargraph.plot(data, [keys, keys], config)

    def hicpro_pairing_chart(self):
        """Generate Pairing chart"""

        # Specify the order of the different possible categories
        keys = {
            "Unique_paired_alignments": {"color": "#005ce6", "name": "Uniquely Aligned"},
            "Low_qual_pairs": {"color": "#b97b35", "name": "Low Quality"},
            "Pairs_with_singleton": {"color": "#ff9933", "name": "Singleton"},
            "Multiple_pairs_alignments": {"color": "#e67300", "name": "Multi Aligned"},
            "Unmapped_airs": {"color": "#a9a2a2", "name": "Failed To Align"},
        }

        # Config for the plot
        config = {
            "id": "hicpro_pairing_stats_plot",
            "title": "HiC-Pro: Pairing Statistics",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
        }

        if not any([k in self.hicpro_data[s_name] for s_name in self.hicpro_data for k in keys]):
            # No data found to build this plot
            return

        return bargraph.plot(self.hicpro_data, keys, config)

    def hicpro_filtering_chart(self):
        """Generate the HiC-Pro filtering plot"""

        # Specify the order of the different possible categories
        keys = {
            "Valid_interaction_pairs_FF": {"color": "#ccddff", "name": "Valid Pairs FF"},
            "Valid_interaction_pairs_RR": {"color": "#6699ff", "name": "Valid Pairs RR"},
            "Valid_interaction_pairs_RF": {"color": "#0055ff", "name": "Valid Pairs RF"},
            "Valid_interaction_pairs_FR": {"color": "#003399", "name": "Valid Pairs FR"},
            "Self_Cycle_pairs": {"color": "#ffad99", "name": "Same Fragment - Self-Circle"},
            "Dangling_end_pairs": {"color": "#ff5c33", "name": "Same Fragment - Dangling Ends"},
            "Religation_pairs": {"color": "#cc2900", "name": "Re-ligation"},
            "Filtered_pairs": {"color": "#661400", "name": "Filtered pairs"},
            "Dumped_pairs": {"color": "#330a00", "name": "Dumped pairs"},
        }

        # Config for the plot
        config = {
            "id": "hicpro_filtering_plot",
            "title": "HiC-Pro: Filtering Statistics",
            "ylab": "# Read Pairs",
            "cpswitch_counts_label": "Number of Read Pairs",
        }

        if not any([k in self.hicpro_data[s_name] for s_name in self.hicpro_data for k in keys]):
            # No data found to build this plot
            return

        return bargraph.plot(self.hicpro_data, keys, config)

    def hicpro_contact_chart(self):
        """Generate the HiC-Pro interaction plot"""

        # Specify the order of the different possible categories
        keys = {
            "cis_shortRange": {"color": "#0039e6", "name": "Unique: cis <= 20Kbp"},
            "cis_longRange": {"color": "#809fff", "name": "Unique: cis > 20Kbp"},
            "trans_interaction": {"color": "#009933", "name": "Unique: trans"},
            "duplicates": {"color": "#a9a2a2", "name": "Duplicate read pairs"},
        }

        # Config for the plot
        config = {
            "id": "hicpro_contact_plot",
            "title": "HiC-Pro: Contact Statistics",
            "ylab": "# Pairs",
            "cpswitch_counts_label": "Number of Pairs",
        }

        if not any([k in self.hicpro_data[s_name] for s_name in self.hicpro_data for k in keys]):
            # No data found to build this plot
            return

        return bargraph.plot(self.hicpro_data, keys, config)

    def hicpro_as_chart(self):
        """Generate Allele-specific plot"""

        keys = {
            "Valid_pairs_from_ref_genome_(1-1)": {"color": "#e6550d", "name": "Genome1 specific read pairs (1-1)"},
            "Valid_pairs_from_ref_genome_with_one_unassigned_mate_(0-1/1-0)": {
                "color": "#fdae6b",
                "name": "Genome1 with one unassigned mate (0-1/1-0)",
            },
            "Valid_pairs_from_alt_genome_(2-2)": {"color": "#756bb1", "name": "Genome2 specific read pairs (2-2)"},
            "Valid_pairs_from_alt_genome_with_one_unassigned_mate_(0-2/2-0)": {
                "color": "#bcbddc",
                "name": "Genome2 with one unassigned mate (0-2/2-0)",
            },
            "Valid_pairs_from_alt_and_ref_genome_(1-2/2-1)": {
                "color": "#a6611a",
                "name": "Trans homologous read pairs (1-2/2/1)",
            },
            "Valid_pairs_with_both_unassigned_mated_(0-0)": {"color": "#cccccc", "name": "Unassigned read pairs"},
            "Valid_pairs_with_at_least_one_conflicting_mate_(3-)": {
                "color": "#a9a2a2",
                "name": "Conflicting read pairs",
            },
        }

        # Config for the plot
        config = {
            "id": "hicpro_asan_plot",
            "title": "HiC-Pro: Allele-specific Statistics",
            "ylab": "# Pairs",
            "cpswitch_counts_label": "Number of Pairs",
        }

        if not any([k in self.hicpro_data[s_name] for s_name in self.hicpro_data for k in keys]):
            # No data found to build this plot
            return

        return bargraph.plot(self.hicpro_data, keys, config)

    def hicpro_capture_chart(self):
        """Generate Capture Hi-C plot"""

        keys = {
            "valid_pairs_on_target_cap_cap": {"color": "#0039e6", "name": "Capture-Capture interactions"},
            "valid_pairs_on_target_cap_rep": {"color": "#809fff", "name": "Capture-Reporter interactions"},
            "valid_pairs_off_target": {"color": "#cccccc", "name": "Off-target valid pairs"},
        }

        # Config for the plot
        config = {
            "id": "hicpro_cap_plot",
            "title": "HiC-Pro: Capture Statistics",
            "ylab": "# Pairs",
            "cpswitch_counts_label": "Number of Pairs",
        }

        if not any([k in self.hicpro_data[s_name] for s_name in self.hicpro_data for k in keys]):
            # No data found to build this plot
            return

        return bargraph.plot(self.hicpro_data, keys, config)



import copy
import logging
from collections import defaultdict

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import linegraph

from .util import average_from_range

log = logging.getLogger(__name__)

N_QV = 2


class DragenBaseMetrics(BaseMultiqcModule):
    """
    Rendering ALL THE THINGS!
    """

    def add_base_metrics(self):
        self.positional_mean_quality_plot()
        return self.dragen_fastqc_data.keys()

    def positional_mean_quality_plot(self):
        """Create the HTML for the positional mean-quality score plot"""

        base_dict = {"A": {}, "C": {}, "G": {}, "T": {}, "N": {}}

        AVG_GROUP = "POSITIONAL BASE MEAN QUALITY"
        COUNT_GROUP = "POSITIONAL BASE CONTENT"

        data = dict()
        for s_name in sorted(self.dragen_fastqc_data):
            for mate in sorted(self.dragen_fastqc_data[s_name]):
                # Parse our per-base, per-position average qualities into a dictionary
                avgs = copy.deepcopy(base_dict)
                for key, value in self.dragen_fastqc_data[s_name][mate][AVG_GROUP].items():
                    if value == "NA":
                        continue
                    parts = key.split()
                    pos = average_from_range(parts[1])
                    base = parts[2].upper()
                    avgs[base][pos] = float(value)

                # Parse matching per-base and total counts by position
                counts = copy.deepcopy(base_dict)
                totals = defaultdict(int)
                for key, value in self.dragen_fastqc_data[s_name][mate][COUNT_GROUP].items():
                    parts = key.split()
                    pos = average_from_range(parts[1])
                    base = parts[2].upper()
                    counts[base][pos] = int(value)
                    totals[pos] += int(value)

                # Use the count and averages to recompute total QVs
                qv_sums = defaultdict(int)
                for base, pos_data in counts.items():
                    for pos, count in pos_data.items():
                        if count == 0:
                            continue
                        elif base == "N":
                            qv_sums[pos] += count * N_QV
                        else:
                            qv_sums[pos] += int(round(count * avgs[base][pos]))

                # Compute the positional, base-agnostic mean QV
                r_name = f"{s_name}_{mate}"
                data[r_name] = dict()
                for pos, qv_sum in qv_sums.items():
                    total = totals[pos]
                    if total > 0:
                        data[r_name][int(pos)] = qv_sum / total

        pconfig = {
            "id": "dragen_fastqc_per_base_sequence_quality_plot",
            "title": "DRAGEN-QC: Per-Position Quality Scores",
            "ylab": "Phred Quality Score",
            "xlab": "Position (bp)",
            "ymin": 0,
            "xDecimals": False,
            "tt_label": "<b>Base {point.x}</b>: {point.y:.2f}",
            # 'colors': self.get_status_cols('per_base_sequence_quality'),
            "yPlotBands": [
                {"from": 28, "to": 100, "color": "#c3e6c3"},
                {"from": 20, "to": 28, "color": "#e6dcc3"},
                {"from": 0, "to": 20, "color": "#e6c3c3"},
            ],
        }

        self.add_section(
            name="Per-Position Mean Quality Scores",
            anchor="dragen_fastqc_per_base_sequence_quality",
            description="The mean quality value across each base position in the read.",
            helptext="""
            To enable multiple samples to be plotted on the same graph, only the mean quality
            scores are plotted (unlike the box plots seen in FastQC reports).

            Taken from the [FastQC help](http://www.bioinformatics.babraham.ac.uk/projects/fastqc/Help/3%20Analysis%20Modules/2%20Per%20Base%20Sequence%20Quality.html):

            _The y-axis on the graph shows the quality scores. The higher the score, the better
            the base call. The background of the graph divides the y axis into very good quality
            calls (green), calls of reasonable quality (orange), and calls of poor quality (red).
            The quality of calls on most platforms will degrade as the run progresses, so it is
            common to see base calls falling into the orange area towards the end of a read._
            """,
            plot=linegraph.plot(data, pconfig),
        )



""" MultiQC module to parse output from STAR """


import logging
import os
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="STAR",
            anchor="star",
            href="https://github.com/alexdobin/STAR",
            info="is an ultrafast universal RNA-seq aligner.",
            doi="10.1093/bioinformatics/bts635",
        )

        # Find and load any STAR reports
        self.star_data = dict()
        for f in self.find_log_files("star"):
            parsed_data = self.parse_star_report(f["f"])
            if parsed_data is not None:
                s_name = f["s_name"]
                if s_name == "" or s_name == "Log.final.out":
                    s_name = self.clean_s_name(os.path.basename(f["root"]), f, root=os.path.dirname(f["root"]))
                if s_name in self.star_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                self.add_data_source(f, section="SummaryLog")
                self.star_data[s_name] = parsed_data

        # Find and load any STAR gene count tables
        self.star_genecounts_unstranded = dict()
        self.star_genecounts_first_strand = dict()
        self.star_genecounts_second_strand = dict()
        for f in self.find_log_files("star/genecounts", filehandles=True):
            parsed_data = self.parse_star_genecount_report(f)
            if parsed_data is not None:
                s_name = f["s_name"]
                if s_name == "" or s_name == "ReadsPerGene.out.tab":
                    s_name = self.clean_s_name(os.path.basename(f["root"]), f, root=os.path.dirname(f["root"]))
                if s_name in self.star_data:
                    log.debug(f"Duplicate ReadsPerGene sample name found! Overwriting: {s_name}")
                self.add_data_source(f, section="ReadsPerGene")
                self.star_genecounts_unstranded[s_name] = parsed_data["unstranded"]
                self.star_genecounts_first_strand[s_name] = parsed_data["first_strand"]
                self.star_genecounts_second_strand[s_name] = parsed_data["second_strand"]

        # Filter to strip out ignored sample names
        self.star_data = self.ignore_samples(self.star_data)
        self.star_genecounts_unstranded = self.ignore_samples(self.star_genecounts_unstranded)
        self.star_genecounts_first_strand = self.ignore_samples(self.star_genecounts_first_strand)
        self.star_genecounts_second_strand = self.ignore_samples(self.star_genecounts_second_strand)

        if len(self.star_data) == 0 and len(self.star_genecounts_unstranded) == 0:
            raise ModuleNoSamplesFound

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        if len(self.star_data) > 0:
            if len(self.star_genecounts_unstranded) > 0:
                log.info(
                    "Found {} reports and {} gene count files".format(
                        len(self.star_data), len(self.star_genecounts_unstranded)
                    )
                )
            else:
                log.info(f"Found {len(self.star_data)} reports")
        else:
            log.info(f"Found {len(self.star_genecounts_unstranded)} gene count files")

        if len(self.star_data) > 0:
            # Write parsed report data to a file
            self.write_data_file(self.star_data, "multiqc_star")

            # Basic Stats Table
            self.star_stats_table()

            # Alignment bar plot
            self.add_section(name="Alignment Scores", anchor="star_alignments", plot=self.star_alignment_chart())

        if len(self.star_genecounts_unstranded) > 0:
            self.add_section(
                name="Gene Counts",
                anchor="star_geneCounts",
                description="Statistics from results generated using <code>--quantMode GeneCounts</code>. "
                + "The three tabs show counts for unstranded RNA-seq, counts for the 1st read strand "
                + "aligned with RNA and counts for the 2nd read strand aligned with RNA.",
                plot=self.star_genecount_chart(),
            )

    def parse_star_report(self, raw_data):
        """Parse the final STAR log file."""

        regexes = {
            "total_reads": r"Number of input reads \|\s+(\d+)",
            "avg_input_read_length": r"Average input read length \|\s+([\d\.]+)",
            "uniquely_mapped": r"Uniquely mapped reads number \|\s+(\d+)",
            "uniquely_mapped_percent": r"Uniquely mapped reads % \|\s+([\d\.]+)",
            "avg_mapped_read_length": r"Average mapped length \|\s+([\d\.]+)",
            "num_splices": r"Number of splices: Total \|\s+(\d+)",
            "num_annotated_splices": r"Number of splices: Annotated \(sjdb\) \|\s+(\d+)",
            "num_GTAG_splices": r"Number of splices: GT/AG \|\s+(\d+)",
            "num_GCAG_splices": r"Number of splices: GC/AG \|\s+(\d+)",
            "num_ATAC_splices": r"Number of splices: AT/AC \|\s+(\d+)",
            "num_noncanonical_splices": r"Number of splices: Non-canonical \|\s+(\d+)",
            "mismatch_rate": r"Mismatch rate per base, % \|\s+([\d\.]+)",
            "deletion_rate": r"Deletion rate per base \|\s+([\d\.]+)",
            "deletion_length": r"Deletion average length \|\s+([\d\.]+)",
            "insertion_rate": r"Insertion rate per base \|\s+([\d\.]+)",
            "insertion_length": r"Insertion average length \|\s+([\d\.]+)",
            "multimapped": r"Number of reads mapped to multiple loci \|\s+(\d+)",
            "multimapped_percent": r"% of reads mapped to multiple loci \|\s+([\d\.]+)",
            "multimapped_toomany": r"Number of reads mapped to too many loci \|\s+(\d+)",
            "multimapped_toomany_percent": r"% of reads mapped to too many loci \|\s+([\d\.]+)",
            "unmapped_mismatches_percent": r"% of reads unmapped: too many mismatches \|\s+([\d\.]+)",
            "unmapped_tooshort_percent": r"% of reads unmapped: too short \|\s+([\d\.]+)",
            "unmapped_other_percent": r"% of reads unmapped: other \|\s+([\d\.]+)",
        }
        parsed_data = {}
        for k, r in regexes.items():
            r_search = re.search(r, raw_data, re.MULTILINE)
            if r_search:
                parsed_data[k] = float(r_search.group(1))
        # Figure out the numbers for unmapped as for some reason only the percentages are given
        try:
            total_mapped = (
                parsed_data["uniquely_mapped"] + parsed_data["multimapped"] + parsed_data["multimapped_toomany"]
            )
            unmapped_count = parsed_data["total_reads"] - total_mapped
            total_unmapped_percent = (
                parsed_data["unmapped_mismatches_percent"]
                + parsed_data["unmapped_tooshort_percent"]
                + parsed_data["unmapped_other_percent"]
            )
            try:
                parsed_data["unmapped_mismatches"] = int(
                    round(unmapped_count * (parsed_data["unmapped_mismatches_percent"] / total_unmapped_percent), 0)
                )
                parsed_data["unmapped_tooshort"] = int(
                    round(unmapped_count * (parsed_data["unmapped_tooshort_percent"] / total_unmapped_percent), 0)
                )
                parsed_data["unmapped_other"] = int(
                    round(unmapped_count * (parsed_data["unmapped_other_percent"] / total_unmapped_percent), 0)
                )
            except ZeroDivisionError:
                parsed_data["unmapped_mismatches"] = 0
                parsed_data["unmapped_tooshort"] = 0
                parsed_data["unmapped_other"] = 0
        except KeyError:
            pass

        if len(parsed_data) == 0:
            return None
        return parsed_data

    def parse_star_genecount_report(self, f):
        """Parse a STAR gene counts output file"""
        # Three numeric columns: unstranded, stranded/first-strand, stranded/second-strand
        keys = ["N_unmapped", "N_multimapping", "N_noFeature", "N_ambiguous"]
        unstranded = {"N_genes": 0}
        first_strand = {"N_genes": 0}
        second_strand = {"N_genes": 0}
        num_errors = 0
        num_genes = 0
        for line in f["f"]:
            s = line.split("\t")
            try:
                for i in [1, 2, 3]:
                    s[i] = float(s[i])
                if s[0] in keys:
                    unstranded[s[0]] = s[1]
                    first_strand[s[0]] = s[2]
                    second_strand[s[0]] = s[3]
                else:
                    unstranded["N_genes"] += s[1]
                    first_strand["N_genes"] += s[2]
                    second_strand["N_genes"] += s[3]
                    num_genes += 1
            except IndexError:
                # Tolerate a few errors in case there is something random added at the top of the file
                num_errors += 1
                if num_errors > 10 and num_genes == 0:
                    log.warning(f"Error parsing {f['fn']}")
                    return None
        if num_genes > 0:
            return {"unstranded": unstranded, "first_strand": first_strand, "second_strand": second_strand}
        else:
            return None

    def star_stats_table(self):
        """Take the parsed stats from the STAR report and add them to the
        basic stats table at the top of the report"""

        headers = {
            "uniquely_mapped_percent": {
                "title": "% Aligned",
                "description": "% Uniquely mapped reads",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn",
            },
            "uniquely_mapped": {
                "title": f"{config.read_count_prefix} Aligned",
                "description": f"Uniquely mapped reads ({config.read_count_desc})",
                "min": 0,
                "scale": "PuRd",
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
            },
        }
        self.general_stats_addcols(self.star_data, headers)

    def star_alignment_chart(self):
        """Make the plot showing alignment rates"""

        # Specify the order of the different possible categories
        keys = {
            "uniquely_mapped": {"color": "#437bb1", "name": "Uniquely mapped"},
            "multimapped": {"color": "#7cb5ec", "name": "Mapped to multiple loci"},
            "multimapped_toomany": {"color": "#f7a35c", "name": "Mapped to too many loci"},
            "unmapped_mismatches": {"color": "#e63491", "name": "Unmapped: too many mismatches"},
            "unmapped_tooshort": {"color": "#b1084c", "name": "Unmapped: too short"},
            "unmapped_other": {"color": "#7f0000", "name": "Unmapped: other"},
        }

        # Config for the plot
        pconfig = {
            "id": "star_alignment_plot",
            "title": "STAR: Alignment Scores",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
        }

        return bargraph.plot(self.star_data, keys, pconfig)

    def star_genecount_chart(self):
        """Make a plot for the ReadsPerGene output"""

        # Specify the order of the different possible categories
        keys = {
            "N_genes": {"color": "#2f7ed8", "name": "Overlapping Genes"},
            "N_noFeature": {"color": "#0d233a", "name": "No Feature"},
            "N_ambiguous": {"color": "#492970", "name": "Ambiguous Features"},
            "N_multimapping": {"color": "#f28f43", "name": "Multimapping"},
            "N_unmapped": {"color": "#7f0000", "name": "Unmapped"},
        }

        # Config for the plot
        pconfig = {
            "id": "star_gene_counts",
            "title": "STAR: Gene Counts",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
            "data_labels": ["Unstranded", "Same Stranded", "Reverse Stranded"],
        }
        datasets = [
            self.star_genecounts_unstranded,
            self.star_genecounts_first_strand,
            self.star_genecounts_second_strand,
        ]
        return bargraph.plot(datasets, [keys, keys, keys, keys], pconfig)



# coding=utf-8


import itertools
import logging
from collections import defaultdict

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import bargraph, table

from .utils import Metric, exist_and_number, make_headers

# Initialise the logger
log = logging.getLogger(__name__)


NAMESPACE = "Mapping"


class DragenMappingMetics(BaseMultiqcModule):
    def add_mapping_metrics(self):
        data_by_rg_by_sample = defaultdict(dict)
        data_by_phenotype_by_sample = defaultdict(dict)

        for f in self.find_log_files("dragen/mapping_metrics"):
            data_by_readgroup, data_by_phenotype = parse_mapping_metrics_file(f)
            s_name = f["s_name"]
            if s_name in data_by_rg_by_sample:
                log.debug(f"Duplicate DRAGEN output prefix found! Overwriting: {s_name}")
            self.add_data_source(f, section="mapping_metrics")
            data_by_phenotype_by_sample[s_name].update(data_by_phenotype)

            for phenotype, phenotype_d in data_by_readgroup.items():
                for rg, data in phenotype_d.items():
                    if any(rg in d_rg for sn, d_rg in data_by_rg_by_sample.items()):
                        log.debug(f"Duplicate read group name {rg} found for output prefix {s_name}! Overwriting")
            data_by_rg_by_sample[s_name].update(data_by_readgroup)

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, s_name)

        # filter to strip out ignored sample names:
        data_by_rg_by_sample = self.ignore_samples(data_by_rg_by_sample)
        data_by_phenotype_by_sample = self.ignore_samples(data_by_phenotype_by_sample)

        # flattening phenotype-sample data by adding a prefix " normal" to the normal samples
        data_by_sample = dict()
        for sn in data_by_phenotype_by_sample:
            for phenotype in data_by_phenotype_by_sample[sn]:
                new_sn = sn
                if phenotype == "normal":
                    new_sn = sn + "_normal"
                data_by_sample[new_sn] = data_by_phenotype_by_sample[sn][phenotype]

        # flattening phenotype-sample data by adding a prefix " normal" to the normal samples
        data_by_rg_by_sample_new = dict()
        for sn in data_by_rg_by_sample:
            for phenotype in data_by_rg_by_sample[sn]:
                new_sn = sn
                if phenotype == "normal":
                    new_sn = sn + "_normal"
                data_by_rg_by_sample_new[new_sn] = data_by_rg_by_sample[sn][phenotype]
        data_by_rg_by_sample = data_by_rg_by_sample_new
        del data_by_rg_by_sample_new

        if not data_by_rg_by_sample and not data_by_phenotype_by_sample:
            return set()

        # Write data to file
        self.write_data_file(data_by_sample, "dragen_map_metrics")

        self.report_mapping_metrics(data_by_sample, data_by_rg_by_sample)

        return data_by_rg_by_sample.keys()

    def report_mapping_metrics(self, data_by_sample, data_by_rg_by_sample):
        # merging all read group data
        data_by_rg = dict()

        for sname in data_by_rg_by_sample:
            for rg, d in data_by_rg_by_sample[sname].items():
                # Check only one read-group per sample
                if len(list(data_by_rg_by_sample[sname].keys())) == 1:
                    rg = sname
                    data_by_rg[rg] = d
                    continue

                # Multiple read-groups per sample
                rg = sname + f" ({rg}) "
                data_by_rg[rg] = d

        # getting all available metric names to determine table headers
        all_metric_names = set()
        for sn, d_by_rg in data_by_rg_by_sample.items():
            for rg, data in d_by_rg.items():
                for m in data.keys():
                    all_metric_names.add(m)
        # and making headers
        genstats_headers, own_tabl_headers = make_headers(all_metric_names, MAPPING_METRICS)

        self.general_stats_addcols(data_by_sample, genstats_headers, namespace=NAMESPACE)

        self.add_section(
            name="Mapping metrics",
            anchor="dragen-mapping-metrics",
            description="""
            Mapping metrics, similar to the metrics computed by the samtools-stats command.
            Shown on per read group level. To see per-sample level metrics, refer to the general
            stats table.
            """,
            plot=table.plot(
                data_by_rg,
                own_tabl_headers,
                pconfig={
                    "id": "dragen-mapping-metrics-table",
                    "namespace": NAMESPACE,
                    "title": "DRAGEN: Mapping metrics",
                },
            ),
        )

        # Skip adding the barplot if it's not informative, such as if all
        #  reads are unmapped due to a FastQcOnly workflow
        all_unmapped = True
        unmapped_key = "Unmapped reads pct"
        for rg, data in data_by_rg.items():
            if unmapped_key in data and data[unmapped_key] < 100.0:
                all_unmapped = False

        if all_unmapped:
            return set()

        # Make bargraph plots of mapped, dupped and paired reads
        self.__map_pair_dup_read_chart(data_by_rg)

    def __map_pair_dup_read_chart(self, data_by_sample):
        paired_reads_data = {}
        mapped_reads_data = {}
        category_labels = []
        data_labels = []

        add_paired_label = False
        add_mapped_label = False
        for sample_id, data in data_by_sample.items():
            # Dragen 3.9 has replaced 'rRNA filtered reads' with 'Adjustment of reads matching filter contigs'
            if "rRNA filtered reads" in data.keys():
                rrna_filtered_reads_key = "rRNA filtered reads"
            elif "Adjustment of reads matching filter contigs" in data.keys():
                rrna_filtered_reads_key = "Adjustment of reads matching filter contigs"
            else:
                rrna_filtered_reads_key = None

            if data.get("Mapped reads R2", None) == 0:
                log.debug(f"single-ended data detected, skipping mapping/paired percentages plot for: {sample_id}")
            elif data.get("Not properly paired reads (discordant)", 0) + data.get(
                "Properly paired reads", 0
            ) + data.get("Singleton reads (itself mapped; mate unmapped)", 0) + data.get("Unmapped reads", 0) + (
                data.get(rrna_filtered_reads_key, 0)
                if rrna_filtered_reads_key is not None and rrna_filtered_reads_key == "rRNA filtered reads"
                else 0
            ) != data.get("Total reads in RG", 0):
                log.warning(
                    "sum of unpaired/discordant/proppaired/unmapped reads not matching total, "
                    "skipping mapping/paired percentages plot for: {}".format(sample_id)
                )
            else:
                paired_reads_data[sample_id] = data
                add_paired_label = True

            if data.get("Number of unique & mapped reads (excl. duplicate marked reads)", 0) + data.get(
                "Number of duplicate marked reads", 0
            ) + data.get("Unmapped reads", 0) + (
                data.get(rrna_filtered_reads_key, 0)
                if rrna_filtered_reads_key is not None and rrna_filtered_reads_key == "rRNA filtered reads"
                else 0
            ) != data.get("Total reads in RG", 0):
                log.warning(
                    "sum of unique/duplicate/unmapped reads not matching total, "
                    "skipping mapping/duplicates percentages plot for: {}".format(sample_id)
                )
            else:
                mapped_reads_data[sample_id] = data
                add_mapped_label = True

        # Add labels
        if add_paired_label:
            category_labels.append(
                {
                    "Properly paired reads": {"color": "#099109", "name": "Paired, properly"},
                    "Not properly paired reads (discordant)": {"color": "#c27a0e", "name": "Paired, discordant"},
                    "Singleton reads (itself mapped; mate unmapped)": {"color": "#912476", "name": "Singleton"},
                    "Unmapped reads": {"color": "#b1084c", "name": "Unmapped"},
                }
            )
            data_labels.append(
                {"name": "Paired vs. discordant vs. singleton", "ylab": "Reads", "cpswitch_counts_label": "Reads"}
            )
        if add_mapped_label:
            mapped_chart_labels = {
                "Number of unique & mapped reads (excl. duplicate marked reads)": {
                    "color": "#437bb1",
                    "name": "Unique",
                },
                "Number of duplicate marked reads": {"color": "#f5a742", "name": "Duplicated"},
                "Unmapped reads": {"color": "#b1084c", "name": "Unmapped"},
            }
            if "Adjustment of reads matching filter contigs" in next(iter(data_by_sample.values())):
                mapped_chart_labels["Adjustment of reads matching filter contigs"] = {
                    "color": "#43b14a",
                    "name": "rRNA filtered",
                }
            elif "rRNA filtered reads" in next(iter(data_by_sample.values())):
                mapped_chart_labels["rRNA filtered reads"] = {"color": "#43b14a", "name": "rRNA filtered"}
            elif "Adjustment of reads matching filter contigs" in next(iter(data_by_sample.values())):
                mapped_chart_labels["Adjustment of reads matching filter contigs"] = {
                    "color": "#43b14a",
                    "name": "rRNA filtered",
                }
            category_labels.append(mapped_chart_labels)
            data_labels.append(
                {"name": "Unique vs duplicated vs unmapped", "ylab": "Reads", "cpswitch_counts_label": "Reads"}
            )

        data_to_plot = [d for d in [paired_reads_data, mapped_reads_data] if d]  # Leaves out empty dicts
        self.add_section(
            name="Mapped / paired / duplicated",
            anchor="dragen-mapped-paired-duplicated",
            description="Distribution of reads based on pairing, duplication and mapping.",
            plot=bargraph.plot(
                data_to_plot,
                category_labels,
                {
                    "id": "mapping_dup_percentage_plot",
                    "title": "Dragen: Mapped/paired/duplicated reads per read group",
                    "ylab": "Reads",
                    "cpswitch_counts_label": "Reads",
                    "data_labels": data_labels,
                },
            ),
        )


def parse_mapping_metrics_file(f):
    """
    Mapping and aligning metrics, like the metrics computed by the Samtools Flagstat command, are available
    on an aggregate level (over all input data), and on a per read group level. Unless explicitly stated,
    the metrics units are in reads (ie, not in terms of pairs or alignments).

    T_SRR7890936_50pc.mapping_metrics.csv

    # phenotype-level metrics (tumor or normal):
    TUMOR MAPPING/ALIGNING SUMMARY,,Total input reads,2200000000,100.00
    TUMOR MAPPING/ALIGNING SUMMARY,,Number of duplicate marked reads,433637413,19.71
    TUMOR MAPPING/ALIGNING SUMMARY,,Number of duplicate marked and mate reads removed,NA
    TUMOR MAPPING/ALIGNING SUMMARY,,Number of unique reads (excl. duplicate marked reads),1766362587,80.29
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with mate sequenced,2200000000,100.00
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads without mate sequenced,0,0.00
    TUMOR MAPPING/ALIGNING SUMMARY,,QC-failed reads,0,0.00
    TUMOR MAPPING/ALIGNING SUMMARY,,Mapped reads,2130883930,96.86
    TUMOR MAPPING/ALIGNING SUMMARY,,Mapped reads R1,1066701794,96.97
    TUMOR MAPPING/ALIGNING SUMMARY,,Mapped reads R2,1064182136,96.74
    TUMOR MAPPING/ALIGNING SUMMARY,,Number of unique & mapped reads (excl. duplicate marked reads),1697246517,77.15
    TUMOR MAPPING/ALIGNING SUMMARY,,Unmapped reads,69116070,3.14
    TUMOR MAPPING/ALIGNING SUMMARY,,Singleton reads (itself mapped; mate unmapped),3917092,0.18
    TUMOR MAPPING/ALIGNING SUMMARY,,Paired reads (itself & mate mapped),2126966838,96.68
    TUMOR MAPPING/ALIGNING SUMMARY,,Properly paired reads,2103060370,95.59
    TUMOR MAPPING/ALIGNING SUMMARY,,Not properly paired reads (discordant),23906468,1.09
    TUMOR MAPPING/ALIGNING SUMMARY,,Paired reads mapped to different chromosomes,17454370,0.82
    TUMOR MAPPING/ALIGNING SUMMARY,,Paired reads mapped to different chromosomes (MAPQ>=10),6463547,0.30
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with MAPQ [40:inf),2002661377,91.03
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with MAPQ [30:40),7169392,0.33
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with MAPQ [20:30),16644390,0.76
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with MAPQ [10:20),20280057,0.92
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with MAPQ [ 0:10),84128714,3.82
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with MAPQ NA (Unmapped reads),69116070,3.14
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with indel R1,26849051,2.52
    TUMOR MAPPING/ALIGNING SUMMARY,,Reads with indel R2,24810803,2.33
    TUMOR MAPPING/ALIGNING SUMMARY,,Total bases,330000000000
    TUMOR MAPPING/ALIGNING SUMMARY,,Total bases R1,165000000000
    TUMOR MAPPING/ALIGNING SUMMARY,,Total bases R2,165000000000
    TUMOR MAPPING/ALIGNING SUMMARY,,Mapped bases R1,160005269100
    TUMOR MAPPING/ALIGNING SUMMARY,,Mapped bases R2,159627320400
    TUMOR MAPPING/ALIGNING SUMMARY,,Soft-clipped bases R1,1757128997,1.10
    TUMOR MAPPING/ALIGNING SUMMARY,,Soft-clipped bases R2,3208748350,2.01
    TUMOR MAPPING/ALIGNING SUMMARY,,Mismatched bases R1,585802788,0.37
    TUMOR MAPPING/ALIGNING SUMMARY,,Mismatched bases R2,1155805091,0.72
    TUMOR MAPPING/ALIGNING SUMMARY,,Mismatched bases R1 (excl. indels),501394281,0.31
    TUMOR MAPPING/ALIGNING SUMMARY,,Mismatched bases R2 (excl. indels),1073788605,0.67
    TUMOR MAPPING/ALIGNING SUMMARY,,Q30 bases,297564555927,90.17
    TUMOR MAPPING/ALIGNING SUMMARY,,Q30 bases R1,155492239719,94.24
    TUMOR MAPPING/ALIGNING SUMMARY,,Q30 bases R2,142072316208,86.10
    TUMOR MAPPING/ALIGNING SUMMARY,,Q30 bases (excl. dups & clipped bases),246555769158
    TUMOR MAPPING/ALIGNING SUMMARY,,Total alignments,2190085267
    TUMOR MAPPING/ALIGNING SUMMARY,,Secondary alignments,0
    TUMOR MAPPING/ALIGNING SUMMARY,,Supplementary (chimeric) alignments,59201337
    TUMOR MAPPING/ALIGNING SUMMARY,,Estimated read length,150.00
    TUMOR MAPPING/ALIGNING SUMMARY,,Average sequenced coverage over genome,102.83
    TUMOR MAPPING/ALIGNING SUMMARY,,Insert length: mean,383.00
    TUMOR MAPPING/ALIGNING SUMMARY,,Insert length: median,376.00
    TUMOR MAPPING/ALIGNING SUMMARY,,Insert length: standard deviation,85.15
    # general metrics - reporting in the header (except for DRAGEN mapping rate may be different to T and N:
    TUMOR MAPPING/ALIGNING SUMMARY,,Bases in reference genome,3209286105
    TUMOR MAPPING/ALIGNING SUMMARY,,Bases in target bed [% of genome],NA
    TUMOR MAPPING/ALIGNING SUMMARY,,Provided sex chromosome ploidy,NA
    TUMOR MAPPING/ALIGNING SUMMARY,,DRAGEN mapping rate [mil. reads/second],0.39
    # then same for normal:
    NORMAL MAPPING/ALIGNING SUMMARY,,Total input reads,1100000000,100.00
    NORMAL MAPPING/ALIGNING SUMMARY,,Number of duplicate marked reads,123518125,11.23
    ...
    # then tumor and normal per-read-group metrics::
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Total reads in RG,2200000000,100.00
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Number of duplicate marked reads,433637413,19.71
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Number of duplicate marked and mate reads removed,NA
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Number of unique reads (excl. duplicate marked reads),1766362587,80.29
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with mate sequenced,2200000000,100.00
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads without mate sequenced,0,0.00
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,QC-failed reads,0,0.00
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mapped reads,2130883930,96.86
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mapped reads R1,1066701794,96.97
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mapped reads R2,1064182136,96.74
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Number of unique & mapped reads (excl. duplicate marked reads),1697246517,77.15
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Unmapped reads,69116070,3.14
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Singleton reads (itself mapped; mate unmapped),3917092,0.18
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Paired reads (itself & mate mapped),2126966838,96.68
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Properly paired reads,2103060370,95.59
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Not properly paired reads (discordant),23906468,1.09
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Paired reads mapped to different chromosomes,17454370,0.82
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Paired reads mapped to different chromosomes (MAPQ>=10),6463547,0.30
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with MAPQ [40:inf),2002661377,91.03
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with MAPQ [30:40),7169392,0.33
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with MAPQ [20:30),16644390,0.76
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with MAPQ [10:20),20280057,0.92
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with MAPQ [ 0:10),84128714,3.82
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with MAPQ NA (Unmapped reads),69116070,3.14
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with indel R1,26849051,2.52
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Reads with indel R2,24810803,2.33
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Total bases,330000000000
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Total bases R1,165000000000
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Total bases R2,165000000000
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mapped bases R1,160005269100
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mapped bases R2,159627320400
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Soft-clipped bases R1,1757128997,1.10
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Soft-clipped bases R2,3208748350,2.01
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mismatched bases R1,585802788,0.37
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mismatched bases R2,1155805091,0.72
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mismatched bases R1 (excl. indels),501394281,0.31
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Mismatched bases R2 (excl. indels),1073788605,0.67
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Q30 bases,297564555927,90.17
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Q30 bases R1,155492239719,94.24
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Q30 bases R2,142072316208,86.10
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Q30 bases (excl. dups & clipped bases),246555769158
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Total alignments,2190085267
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Secondary alignments,0
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Supplementary (chimeric) alignments,59201337
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Estimated read length,150.00
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Average sequenced coverage over genome,102.83
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Insert length: mean,383.01
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Insert length: median,376.00
    TUMOR MAPPING/ALIGNING PER RG,T_SRR7890936_50pc,Insert length: standard deviation,87.58
    # same for normal:
    NORMAL MAPPING/ALIGNING PER RG,N_SRR7890889,Total reads in RG,1100000000,100.00
    NORMAL MAPPING/ALIGNING PER RG,N_SRR7890889,Number of duplicate marked reads,NA
    NORMAL MAPPING/ALIGNING PER RG,N_SRR7890889,Number of duplicate marked and mate reads removed,NA
    NORMAL MAPPING/ALIGNING PER RG,N_SRR7890889,Number of unique reads (excl. duplicate marked reads),NA
    ...

    We are reporting summary metrics in the general stats table, and per-read-group in a separate table.
    """

    data_by_readgroup = defaultdict(dict)
    data_by_phenotype = defaultdict(dict)

    for line in f["f"].splitlines():
        fields = line.split(",")
        phenotype = fields[0].split("/")[0].split(" ")[0].lower()  # TUMOR MAPPING -> tumor
        analysis = fields[0].split("/")[1]  # ALIGNING SUMMARY, ALIGNING PER RG
        metric = fields[2]
        value = fields[3] if fields[3] != "NA" else None
        try:
            value = int(value)
        except (ValueError, TypeError):
            try:
                value = float(value)
            except (ValueError, TypeError):
                pass

        percentage = None
        if len(fields) > 4:  # percentage
            percentage = fields[4]
            try:
                percentage = float(percentage)
            except ValueError:
                pass

        # sample-unspecific metrics are reported only in ALIGNING SUMMARY sections
        if analysis == "ALIGNING SUMMARY":
            if value is not None:
                data_by_phenotype[phenotype][metric] = value
            if percentage is not None:
                data_by_phenotype[phenotype][metric + " pct"] = percentage

        # for sample-specific metrics, using ALIGNING PER RG because it has the sample name in the 2nd col
        if analysis == "ALIGNING PER RG":
            # setting normal and tumor sample names for future use
            readgroup = fields[1]
            if readgroup not in data_by_readgroup[phenotype].keys():
                data_by_readgroup[phenotype][readgroup] = {}
            if value is not None:
                data_by_readgroup[phenotype][readgroup][metric] = value
            if percentage is not None:
                data_by_readgroup[phenotype][readgroup][metric + " pct"] = percentage

    # adding some missing values that we wanna report for consistency
    # Expand data_by_readgroup to values below phenotype level
    for data in itertools.chain(
        *[data_by_readgroup[key].values() for key in data_by_readgroup.keys()], data_by_phenotype.values()
    ):
        # fixing when deduplication wasn't performed, or running with single-end data
        for field in [
            "Number of duplicate marked and mate reads removed",
            "Number of unique reads (excl. duplicate marked reads)",
            "Mismatched bases R2 (excl. indels)",
            "Mismatched bases R2",
            "Soft-clipped bases R2",
        ]:
            try:
                data.pop(field)
            except KeyError:
                pass

        # adding alignment percentages
        if exist_and_number(data, "Total alignments", "Secondary alignments") and data["Total alignments"] > 0:
            data["Secondary alignments pct"] = data["Secondary alignments"] / data["Total alignments"] * 100.0

        # adding some missing bases percentages
        if exist_and_number(data, "Total bases") and data["Total bases"] > 0:
            for m in ["Q30 bases (excl. dups & clipped bases)", "Mapped bases R1", "Mapped bases R2"]:
                if exist_and_number(data, m):
                    data[m + " pct"] = data[m] / data["Total bases"] * 100.0

    return data_by_readgroup, data_by_phenotype


MAPPING_METRICS = [
    # id_in_data                                                    # title (display name)  # in_genstats  # in_own_tabl # unit  # description
    # Read stats:
    Metric(
        "Total input reads",
        "Input reads",
        "#",
        None,
        "reads",
        "Total number of input reads for this sample (or total number of reads in all input read groups combined), {}",
    ),
    Metric("Total reads in RG", "Input reads", None, "#", "reads", "Total number of reads in this RG, {}"),
    Metric(
        "Average sequenced coverage over genome",
        "Raw depth",
        "hid",
        "#",
        "x",
        "Average sequenced coverage over genome, based on all input reads (including duplicate, clipped and low quality bases and reads)",
    ),
    Metric("Reads with mate sequenced", "Paired", "hid", "%", "reads", "Number of reads with a mate sequenced, {}"),
    Metric(
        "Reads without mate sequenced",
        "Single",
        None,
        "hid",
        "reads",
        "Number of reads without a mate sequenced, {}",
        the_higher_the_worse=True,
    ),
    Metric(
        "QC-failed reads",
        "QC-fail",
        "hid",
        "%",
        "reads",
        "Number of reads not passing platform/vendor quality checks (SAM flag 0x200), {}",
        precision=2,
        the_higher_the_worse=True,
    ),
    Metric("Mapped reads", "Map", "hid", "hid", "reads", "Number of mapped reads, {}"),
    Metric("Mapped reads R1", "Map R1", None, "hid", "reads", "Number of mapped reads R1, {}"),
    Metric("Mapped reads R2", "Map R2", None, "hid", "reads", "Number of mapped reads R2, {}"),
    Metric("Unmapped reads", "Unmap", "%", "%", "reads", "Number of unmapped reads, {}", the_higher_the_worse=True),
    Metric(
        "rRNA filtered reads", "rRNA", "%", "%", "reads", "Number of rRNA filtered reads, {}", the_higher_the_worse=True
    ),
    Metric("Reads with MAPQ [40:inf)", "MQ⩾40", None, "hid", "reads", "Number of reads with MAPQ [40:inf), {}"),
    Metric(
        "Number of duplicate marked reads",
        "Dup",
        "%",
        "%",
        "reads",
        "Number of duplicate marked reads as a result of the --enable-duplicatemarking option being used, {}",
        the_higher_the_worse=True,
    ),
    Metric(
        "Number of duplicate marked and mate reads removed",
        "Dup with mates removed",
        None,
        "hid",
        "reads",
        "Number of reads marked as duplicates, along with any mate reads, that are removed "
        "when the --remove-duplicates option is used, {}",
        the_higher_the_worse=True,
    ),
    Metric(
        "Number of unique reads (excl. duplicate marked reads)",
        "Uniq",
        None,
        "hid",
        "reads",
        "Number of unique reads (all reads minus duplicate marked reads), {}",
    ),
    Metric(
        "Number of unique & mapped reads (excl. duplicate marked reads)",
        "Uniq map",
        "hid",
        "hid",
        "reads",
        "Number of unique & mapped reads (mapped reads minus duplicate marked reads), {}",
    ),
    Metric(
        "Paired reads (itself & mate mapped)",
        "Self & mate map",
        None,
        "hid",
        "reads",
        "Number of reads mapped in pairs (itself & mate mapped), {}",
    ),
    Metric(
        "Properly paired reads",
        "Prop pair",
        "%",
        "%",
        "reads",
        "Number of properly paired reads, {} (both reads in pair are mapped and "
        "fall within an acceptable range from each other based on the estimated insert length distribution). "
        "Duplicate and low quality reads are not excluded, i.e. the percentage is calculated relative to the total number of reads.",
    ),
    Metric(
        "Not properly paired reads (discordant)",
        "Discord",
        "hid",
        "%",
        "reads",
        "Number of discordant reads: paired reads minus properly paired reads , {}",
        precision=2,
        the_higher_the_worse=True,
    ),
    Metric(
        "Singleton reads (itself mapped; mate unmapped)",
        "Singleton",
        None,
        "%",
        "reads",
        "Number of singleton reads: itself mapped; mate unmapped, {}",
        precision=2,
        the_higher_the_worse=True,
    ),
    Metric(
        "Paired reads mapped to different chromosomes",
        "Diff chrom",
        None,
        "hid",
        "reads",
        "Number of paired reads with a mate mapped to a different chromosome, {}",
        precision=2,
        the_higher_the_worse=True,
    ),
    Metric(
        "Paired reads mapped to different chromosomes (MAPQ>=10)",
        "Diff chr, MQ⩾10",
        "hid",
        "%",
        "reads",
        "Number of paired reads, mapped with MAPQ>=10 and with a mate mapped to a different chromosome, {}",
        precision=2,
        the_higher_the_worse=True,
    ),
    Metric(
        "Reads with indel R1",
        "Reads with indel R1",
        None,
        "hid",
        "reads",
        "Number of R1 reads containing at least 1 indel, {}",
        the_higher_the_worse=True,
    ),
    Metric(
        "Reads with indel R2",
        "Reads with indel R2",
        None,
        "hid",
        "reads",
        "Number of R2 reads containing at least 1 indel, {}",
        the_higher_the_worse=True,
    ),
    # Read length stats
    Metric(
        "Estimated read length",
        "Read len",
        "hid",
        "hid",
        "bp",
        "Estimated read length. Total number of input bases divided by the number of reads",
    ),
    Metric("Insert length: mean", "Avg IS", "hid", "hid", "bp", "Mean insert size"),
    Metric("Insert length: median", "Med IS", "#", "#", "bp", "Median insert size", precision=0),
    Metric(
        "Insert length: standard deviation", "IS std", "hid", "hid", "bp", "Standard deviation of insert size deviation"
    ),
    # Bases stats:
    Metric("Total bases", "Input bases", "hid", "hid", "bases", "Total number of bases sequenced, {}"),
    Metric("Total bases R1", "Input bases R1", None, "hid", "bases", "Total number of bases sequenced on R1 reads, {}"),
    Metric("Total bases R2", "Input bases R2", None, "hid", "bases", "Total number of bases sequenced on R2 reads, {}"),
    Metric("Mapped bases R1", "Mapped bases R1", None, "hid", "bases", "Number of mapped bases on R1 reads, {}"),
    Metric("Mapped bases R2", "Mapped bases R2", None, "hid", "bases", "Number of mapped bases on R2 reads, {}"),
    Metric(
        "Soft-clipped bases R1",
        "Soft-clip bases R1",
        None,
        "hid",
        "bases",
        "Number of soft-clipped bases on R1 reads, {}",
        the_higher_the_worse=True,
    ),
    Metric(
        "Soft-clipped bases R2",
        "Soft-clip bases R2",
        None,
        "hid",
        "bases",
        "Number of soft-clipped bases on R2 reads, {}",
        the_higher_the_worse=True,
    ),
    Metric(
        "Mismatched bases R1",
        "MM bases R1",
        None,
        "hid",
        "bases",
        "Number of mismatched bases on R1, {}, which is the sum of SNP count and indel lengths. "
        "It does not count anything within soft clipping, or RNA introns. "
        "It also does not count a mismatch if either the reference base or read base is N",
        the_higher_the_worse=True,
    ),
    Metric(
        "Mismatched bases R2",
        "MM bases R2",
        None,
        "hid",
        "bases",
        "Number of mismatched bases on R2, {}, which is the sum of SNP count and indel lengths. "
        "It does not count anything within soft clipping, or RNA introns. "
        "It also does not count a mismatch if either the reference base or read base is N",
        the_higher_the_worse=True,
    ),
    Metric(
        "Mismatched bases R1 (excl. indels)",
        "MM bases R1 excl indels",
        None,
        "hid",
        "bases",
        "Number of mismatched bases on R1, {}. The indels lengts are ignored. "
        "It does not count anything within soft clipping, or RNA introns. "
        "It also does not count a mismatch if either the reference base or read base is N",
        the_higher_the_worse=True,
    ),
    Metric(
        "Mismatched bases R2 (excl. indels)",
        "MM bases R2 excl indels",
        None,
        "hid",
        "bases",
        "Number of mismatched bases on R2, {}. The indels lengts are ignored. "
        "It does not count anything within soft clipping, or RNA introns. "
        "It also does not count a mismatch if either the reference base or read base is N",
        the_higher_the_worse=True,
    ),
    Metric("Q30 bases", "Q30", "hid", "hid", "bases", "Number of raw bases with BQ >= 30, {}"),
    Metric("Q30 bases R1", "Q30 R1", None, "hid", "bases", "Number of raw bases on R1 reads with BQ >= 30, {}"),
    Metric("Q30 bases R2", "Q30 R2", None, "hid", "bases", "Number of raw bases on R2 reads with BQ >= 30, {}"),
    Metric(
        "Q30 bases (excl. dups & clipped bases)",
        "Q30 excl dup & clipped",
        None,
        "hid",
        "bases",
        "Number of non-clipped bases with BQ >= 30 on non-duplicate reads, {}",
    ),
    # Alignments stats:
    Metric("Total alignments", "Alignments", "hid", "#", "reads", "Total number of alignments with MQ > 0, {}"),
    Metric(
        "Secondary alignments",
        "Sec'ry",
        "hid",
        "%",
        "reads",
        "Number of secondary alignments, {}. Secondary alignment occurs when "
        "a given read could align reasonably well to more than one place. "
        'One of the possible reported alignments is termed "primary" and '
        'the others will be marked as "secondary".',
        precision=2,
    ),
    Metric(
        "Supplementary (chimeric) alignments",
        "Suppl'ry",
        "hid",
        "%",
        "reads",
        "Number of supplementary (chimeric) alignments, {}. A chimeric read is split "
        "over multiple loci (possibly due to structural variants). One alignment is "
        "referred to as the representative alignment, the other are supplementary",
        precision=2,
    ),
    Metric(
        "Estimated sample contamination",
        "Contam'n",
        "#",
        "#",
        "proportion",
        "Fraction of estimated sample contamination",
        precision=2,
        the_higher_the_worse=True,
    ),
    Metric(
        "rRNA filtered reads", "rRNA", "%", "%", "reads", "Number of rRNA filtered reads, {}", the_higher_the_worse=True
    ),
    Metric(
        "Adjustment of reads matching filter contigs",
        "rRNA / Filtered Contigs",
        "%",
        "%",
        "reads",
        "Number of filtered reads, {}",
        the_higher_the_worse=True,
    ),
]



""" MultiQC module to parse output from Bowtie 1 """


import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """Bowtie 1 module, parses stderr logs."""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Bowtie 1",
            anchor="bowtie1",
            target="Bowtie 1",
            href="http://bowtie-bio.sourceforge.net/",
            info="is an ultrafast, memory-efficient short read aligner.",
            doi="10.1186/gb-2009-10-3-r25",
        )

        # Find and load any Bowtie reports
        self.bowtie_data = dict()
        for f in self.find_log_files("bowtie1"):
            self.parse_bowtie_logs(f)

        # Filter to strip out ignored sample names
        self.bowtie_data = self.ignore_samples(self.bowtie_data)

        if len(self.bowtie_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.bowtie_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.bowtie_data, "multiqc_bowtie1")

        # Basic Stats Table
        # Report table is immutable, so just updating it works
        self.bowtie_general_stats_table()

        # Alignment Rate Plot
        self.bowtie_alignment_plot()

    def parse_bowtie_logs(self, f):
        s_name = f["s_name"]
        parsed_data = {}
        regexes = {
            "reads_processed": r"# reads processed:\s+(\d+)",
            "reads_aligned": r"# reads with at least one(?: reported)? alignment:\s+(\d+)",
            "reads_aligned_percentage": r"# reads with at least one(?: reported)? alignment:\s+\d+\s+\(([\d\.]+)%\)",
            "not_aligned": r"# reads that failed to align:\s+(\d+)",
            "not_aligned_percentage": r"# reads that failed to align:\s+\d+\s+\(([\d\.]+)%\)",
            "multimapped": r"# reads with alignments suppressed due to -m:\s+(\d+)",
            "multimapped_percentage": r"# reads with alignments suppressed due to -m:\s+\d+\s+\(([\d\.]+)%\)",
        }

        for line in f["f"].splitlines():
            # Attempt in vain to find original bowtie1 command, logged by another program
            if "bowtie" in line and "q.gz" in line:
                fqmatch = re.search(r"([^\s,]+\.f(ast)?q.gz)", line)
                if fqmatch:
                    s_name = self.clean_s_name(fqmatch.group(1), f)
                    log.debug(f"Found a bowtie command, updating sample name to '{s_name}'")

            # End of log, reset in case there is another in this file
            if "Overall time:" in line:
                if len(parsed_data) > 0:
                    if s_name in self.bowtie_data:
                        log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                    self.add_data_source(f, s_name)
                    self.bowtie_data[s_name] = parsed_data
                s_name = f["s_name"]
                parsed_data = {}

            for k, r in regexes.items():
                match = re.search(r, line)
                if match:
                    parsed_data[k] = float(match.group(1).replace(",", ""))

        if len(parsed_data) > 0:
            if s_name in self.bowtie_data:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.add_data_source(f, s_name)
            self.bowtie_data[s_name] = parsed_data

    def bowtie_general_stats_table(self):
        """Take the parsed stats from the Bowtie report and add it to the
        basic stats table at the top of the report"""

        headers = {
            "reads_aligned_percentage": {
                "title": "% Aligned",
                "description": "% reads with at least one reported alignment",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn",
            },
            "reads_aligned": {
                "title": f"{config.read_count_prefix} Aligned",
                "description": f"reads with at least one reported alignment ({config.read_count_desc})",
                "min": 0,
                "scale": "PuRd",
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
            },
        }
        self.general_stats_addcols(self.bowtie_data, headers)

    def bowtie_alignment_plot(self):
        """Make the HighCharts HTML to plot the alignment rates"""

        # Specify the order of the different possible categories
        keys = {
            "reads_aligned": {"color": "#8bbc21", "name": "Aligned"},
            "multimapped": {"color": "#2f7ed8", "name": "Multimapped"},
            "not_aligned": {"color": "#0d233a", "name": "Not aligned"},
        }

        # Config for the plot
        config = {
            "id": "bowtie1_alignment",
            "title": "Bowtie 1: Alignment Scores",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
        }

        self.add_section(
            description="This plot shows the number of reads aligning to the reference in different ways.",
            helptext="""
            There are 3 possible types of alignment:
            * **Aligned**: Read has only one occurence in the reference genome.
            * **Multimapped**: Read has multiple occurence.
            * **Not aligned**: Read has no occurence.
            """,
            plot=bargraph.plot(self.bowtie_data, keys, config),
        )



""" MultiQC submodule to parse output from RSeQC geneBody_coverage.py
http://rseqc.sourceforge.net/#genebody-coverage-py """

import logging

from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(self):
    """Find RSeQC gene_body_coverage reports and parse their data"""

    # Set up vars
    self.gene_body_cov_hist_counts = dict()
    self.gene_body_cov_hist_percent = dict()

    # TODO - Do separate parsing step to find skewness values
    # and add these to the general stats table?

    # Go through files and parse data
    for f in self.find_log_files("rseqc/gene_body_coverage"):
        # geneBodyCoverage.py
        # RSeQC >= v2.4
        # NB: Capitilisation
        if f["f"].startswith("Percentile"):
            keys = []
            nrows = 0
            for line in f["f"].splitlines():
                s = line.split("\t")
                # Check that this is the right file type (detection by capitilisation is pretty weak!)
                if len(s) < 3:
                    break
                if len(keys) == 0:
                    keys = s[1:]
                else:
                    nrows += 1
                    s_name = self.clean_s_name(s[0], f)
                    if s_name in self.gene_body_cov_hist_counts:
                        log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                    self.add_data_source(f, s_name, section="gene_body_coverage")
                    self.gene_body_cov_hist_counts[s_name] = dict()
                    for k, var in enumerate(s[1:]):
                        self.gene_body_cov_hist_counts[s_name][int(keys[k])] = float(var)
            if nrows == 0:
                log.warning(f"Empty geneBodyCoverage file found: {f['fn']}")

        # geneBodyCoverage2.py
        #   AND
        # geneBodyCoverage.py
        # RSeQC < v2.4
        if f["f"].startswith("Total reads") or f["f"].startswith("percentile"):
            if f["s_name"] in self.gene_body_cov_hist_counts:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
            self.add_data_source(f, section="gene_body_coverage")
            self.gene_body_cov_hist_counts[f["s_name"]] = dict()
            nrows = 0
            for line in f["f"].splitlines():
                s = line.split("\t")
                # Check that this is the right file type (detection by capitilisation is pretty weak!)
                if len(s) > 2:
                    break
                try:
                    nrows += 1
                    self.gene_body_cov_hist_counts[f["s_name"]][int(s[0])] = float(s[1])
                except (IndexError, ValueError):
                    # Header lines don't use tabs, so get IndexError.
                    # Other unexpected weird stuff will not play nicely with int() and float()
                    pass
            if nrows == 0:
                del self.gene_body_cov_hist_counts[f["s_name"]]
                log.warning(f"Empty geneBodyCoverage file found: {f['fn']}")

    # Filter to strip out ignored sample names
    self.gene_body_cov_hist_counts = self.ignore_samples(self.gene_body_cov_hist_counts)

    if len(self.gene_body_cov_hist_counts) == 0:
        return 0

    # Write data to file
    self.write_data_file(self.gene_body_cov_hist_counts, "rseqc_gene_body_cov")

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    self.add_software_version(None)

    # Make a normalised coverage for plotting using the formula (cov - min_cov) / (max_cov - min_cov)
    for s_name in self.gene_body_cov_hist_counts:
        self.gene_body_cov_hist_percent[s_name] = dict()
        # min_cov and max_cov are required to compute the normalized coverage
        min_cov = min(self.gene_body_cov_hist_counts[s_name].values())
        max_cov = max(self.gene_body_cov_hist_counts[s_name].values())
        for k, v in self.gene_body_cov_hist_counts[s_name].items():
            self.gene_body_cov_hist_percent[s_name][k] = (v - min_cov) / (max_cov - min_cov)

    # Add line graph to section
    pconfig = {
        "id": "rseqc_gene_body_coverage_plot",
        "title": "RSeQC: Gene Body Coverage",
        "ylab": "Coverage",
        "xlab": "Gene Body Percentile (5' -> 3')",
        "xmin": 0,
        "xmax": 100,
        "tt_label": "<strong>{point.x}% from 5'</strong>: {point.y:.2f}",
        "data_labels": [
            {"name": "Percentages", "ylab": "Percentage Coverage"},
            {"name": "Counts", "ylab": "Coverage"},
        ],
    }
    self.add_section(
        name="Gene Body Coverage",
        anchor="rseqc-gene_body_coverage",
        description='<a href="http://rseqc.sourceforge.net/#genebody-coverage-py" target="_blank">Gene Body Coverage</a>'
        " calculates read coverage over gene bodies."
        " This is used to check if reads coverage is uniform and"
        " if there is any 5' or 3' bias.",
        plot=linegraph.plot([self.gene_body_cov_hist_percent, self.gene_body_cov_hist_counts], pconfig),
    )

    # Return number of samples found
    return len(self.gene_body_cov_hist_counts)



"""Plotly bargraph functionality."""
import copy
import dataclasses
import logging
from collections import defaultdict
from typing import Dict, List

import math
import plotly.graph_objects as go
import spectra

from multiqc.plots.plotly import determine_barplot_height
from multiqc.plots.plotly.plot import Plot, PlotType, BaseDataset, split_long_string
from multiqc.utils import util_functions, config

logger = logging.getLogger(__name__)


def plot(
    cats_lists: List[List[Dict]],
    samples_lists: List[List[str]],
    pconfig: Dict,
) -> str:
    """
    Build and add the plot data to the report, return an HTML wrapper.
    :param cats_lists: each dataset is a list of dicts with the keys: {name, color, data},
        where `name` is the category name, `color` is the color of the bar,
        and `data` is a list of values for each sample. Each outer list will
        correspond a separate tab.
    :param samples_lists: list of lists of bar names (that is, sample names). Similarly,
        each outer list will correspond to a separate tab.
    :param pconfig: Plot configuration dictionary
    :return: HTML with JS, ready to be inserted into the page
    """
    p = BarPlot(
        pconfig,
        cats_lists,
        samples_lists,
        max_n_samples=max([len(samples) for samples in samples_lists]),
    )

    from multiqc.utils import report

    return p.add_to_report(report)


class BarPlot(Plot):
    @dataclasses.dataclass
    class Dataset(BaseDataset):
        cats: List[Dict]
        samples: List[str]

        @staticmethod
        def create(
            dataset: BaseDataset,
            cats: List[Dict],
            samples: List[str],
        ) -> "BarPlot.Dataset":
            # Need to reverse samples as the bar plot will show them reversed
            samples = list(reversed(samples))
            for cat in cats:
                # Reverse the data to match the reversed samples
                cat["data"] = list(reversed(cat["data"]))
                if "data_pct" in cat:
                    cat["data_pct"] = list(reversed(cat["data_pct"]))

            # Post-process categories
            for cat in cats:
                # Split long category names
                if "name" not in cat:
                    raise ValueError(f"Bar plot {dataset.plot.id}: missing 'name' key in category")
                cat["name"] = "<br>".join(split_long_string(cat["name"]))

                # Reformat color to be ready to add alpha in Plotly-JS
                color = spectra.html(cat["color"])
                cat["color"] = ",".join([f"{x:.2f}" for x in color.rgb])

                # Check that the number of samples is the same for all categories
                assert len(samples) == len(cat["data"])

            dataset = BarPlot.Dataset(
                **dataset.__dict__,
                cats=cats,
                samples=samples,
            )

            return dataset

        def create_figure(
            self,
            layout: go.Layout,
            is_log=False,
            is_pct=False,
        ) -> go.Figure:
            """
            Create a Plotly figure for a dataset
            """
            fig = go.Figure(layout=layout)

            for cat in self.cats:
                data = cat["data_pct"] if is_pct else cat["data"]

                params = copy.deepcopy(self.trace_params)
                params["marker"]["color"] = f"rgb({cat['color']})"
                fig.add_trace(
                    go.Bar(
                        y=self.samples,
                        x=data,
                        name=cat["name"],
                        **params,
                    ),
                )
            return fig

    def __init__(self, pconfig: Dict, cats_lists: List, samples_lists: List, max_n_samples: int):
        super().__init__(PlotType.BAR, pconfig, len(cats_lists))
        if len(cats_lists) != len(samples_lists):
            raise ValueError("Number of datasets and samples lists do not match")

        self.datasets: List[BarPlot.Dataset] = [
            BarPlot.Dataset.create(d, cats=cats, samples=samples)
            for d, cats, samples in zip(self.datasets, cats_lists, samples_lists)
        ]

        # Set the barmode
        barmode = "relative"  # stacking, but drawing negative values below zero
        if "stacking" in pconfig and (pconfig["stacking"] in ["group", None]):
            barmode = "group"  # side by side

        max_n_cats = max([len(dataset.cats) for dataset in self.datasets])

        # Set height to also be proportional to the number of cats to fit a legend
        HEIGHT_PER_LEGEND_ITEM = 19
        legend_height = HEIGHT_PER_LEGEND_ITEM * max_n_cats

        height = determine_barplot_height(
            max_n_samples=max_n_samples,
            # Group mode puts each category in a separate bar, so need to multiply by the number of categories
            max_bars_in_group=max_n_cats if barmode == "group" else 1,
            legend_height=legend_height,
        )

        self.layout.update(
            height=height,
            showlegend=True,
            barmode=barmode,
            bargroupgap=0,
            bargap=0.2,
            yaxis=dict(
                showgrid=False,
                categoryorder="trace",  # keep sample order
                automargin=True,  # to make sure there is enough space for ticks labels
                title=None,
                hoverformat=self.layout.xaxis.hoverformat,
                ticksuffix=self.layout.xaxis.ticksuffix,
                # Prevent JavaScript from automatically parsing categorical values as numbers:
                type="category",
            ),
            xaxis=dict(
                title=dict(text=self.layout.yaxis.title.text),
                hoverformat=self.layout.yaxis.hoverformat,
                ticksuffix=self.layout.yaxis.ticksuffix,
            ),
            # Re-initiate legend to reset to default legend location on the top right
            legend=go.layout.Legend(
                # We use legend groups with subplots to simulate standard legend interactivity
                # like we had a standard bar graph without subplots. We need to remove the space
                # between the legend groups to make it look like a single legend.
                tracegroupgap=0,
                # Plotly plots the grouped bar graph in a reversed order in respect to
                # the legend, so reversing the legend to match it:
                traceorder="normal" if barmode != "group" else "reversed",
            ),
            hovermode="y unified",
            hoverlabel=dict(
                bgcolor="rgba(255, 255, 255, 0.8)",
                font=dict(color="black"),
            ),
        )

        if getattr(config, "barplot_legend_on_bottom", False):
            self.layout.update(
                legend=go.layout.Legend(
                    orientation="h",
                    x=0.5,
                    xanchor="center",
                    y=-0.5,
                    yanchor="top",
                ),
            )

        for dataset in self.datasets:
            if barmode == "group":
                # max category
                xmax_cnt = max(max(cat["data"][i] for cat in dataset.cats) for i in range(len(dataset.samples)))
                xmin_cnt = min(min(cat["data"][i] for cat in dataset.cats) for i in range(len(dataset.samples)))
            else:
                # max sum of all categories across all samples
                xmax_cnt = max(
                    sum(cat["data"][i] if cat["data"][i] > 0 else 0 for cat in dataset.cats)
                    for i in range(len(dataset.samples))
                )
                xmin_cnt = min(
                    sum(cat["data"][i] if cat["data"][i] < 0 else 0 for cat in dataset.cats)
                    for i in range(len(dataset.samples))
                )

            dataset.layout.update(
                yaxis=dict(
                    title=None,
                    hoverformat=dataset.layout["xaxis"]["hoverformat"],
                    ticksuffix=dataset.layout["xaxis"]["ticksuffix"],
                    autorangeoptions=dataset.layout["xaxis"]["autorangeoptions"],
                ),
                xaxis=dict(
                    title=dict(text=dataset.layout["yaxis"]["title"]["text"]),
                    hoverformat=dataset.layout["yaxis"]["hoverformat"],
                    ticksuffix=dataset.layout["yaxis"]["ticksuffix"],
                    autorangeoptions={
                        "minallowed": xmin_cnt,
                        "maxallowed": xmax_cnt,
                    },
                ),
                showlegend=len(dataset.cats) > 1,
            )
            dataset.trace_params.update(
                orientation="h",
                marker=dict(line=dict(width=0)),
                textposition="inside",
                insidetextanchor="start",
            )
            if "hovertemplate" in dataset.trace_params:
                # %{text} doesn't work for unified hovermode:
                dataset.trace_params["hovertemplate"] = dataset.trace_params["hovertemplate"].replace("%{text}", "")

            if dataset.layout["xaxis"]["hoverformat"] is None:
                if all(all(isinstance(x, float) or math.isnan(x) for x in cat["data"]) for cat in dataset.cats):
                    dataset.layout["xaxis"]["hoverformat"] = ",.2f"
                elif all(all(isinstance(x, int) or math.isnan(x) for x in cat["data"]) for cat in dataset.cats):
                    dataset.layout["xaxis"]["hoverformat"] = ",.0f"

        # Expand data with zeroes if there are fewer values than samples
        for dataset in self.datasets:
            for cat in dataset.cats:
                if len(cat["data"]) < len(dataset.samples):
                    cat["data"].extend([0] * (len(dataset.samples) - len(cat["data"])))

        # Calculate and save percentages
        if self.add_pct_tab:
            for pidx, dataset in enumerate(self.datasets):
                # Count totals for each category
                sums = [0 for _ in dataset.cats[0]["data"]]
                for cat in dataset.cats:
                    for sample_idx, val in enumerate(cat["data"]):
                        if not math.isnan(val):
                            sums[sample_idx] += abs(val)

                # Now, calculate percentages for each category
                for cat in dataset.cats:
                    values = [x for x in cat["data"]]
                    for sample_idx, val in enumerate(values):
                        sum_for_sample = sums[sample_idx]
                        if sum_for_sample == 0:
                            values[sample_idx] = 0
                        else:
                            values[sample_idx] = float(val + 0.0) / float(sum_for_sample) * 100.0
                    cat["data_pct"] = values

                if barmode == "group":
                    # calculating the min percentage range as well because it will be negative for negative values
                    dataset.pct_range["xaxis"]["min"] = min(
                        min(cat["data_pct"][i] for cat in dataset.cats) for i in range(len(dataset.samples))
                    )
                else:
                    dataset.pct_range["xaxis"]["min"] = min(
                        sum(cat["data_pct"][i] if cat["data_pct"][i] < 0 else 0 for cat in dataset.cats)
                        for i in range(len(dataset.samples))
                    )

        if self.add_log_tab:
            # Sorting from small to large so the log switch makes sense
            for dataset in self.datasets:
                dataset.cats.sort(key=lambda x: sum(x["data"]))
                # But reversing the legend so the largest bars are still on the top
                self.layout.legend.traceorder = "reversed"

    @staticmethod
    def axis_controlled_by_switches() -> List[str]:
        """
        Return a list of axis names that are controlled by the log10 scale and percentage
        switch buttons
        """
        return ["xaxis"]

    def save_data_file(self, dataset: Dataset) -> None:
        val_by_cat_by_sample = defaultdict(dict)
        for cat in dataset.cats:
            for d_idx, d_val in enumerate(cat["data"]):
                s_name = dataset.samples[d_idx]
                val_by_cat_by_sample[s_name][cat["name"]] = d_val
        util_functions.write_data_file(val_by_cat_by_sample, dataset.uid)

    @staticmethod
    def tt_label() -> str:
        """Default tooltip label"""
        return "%{meta}: <b>%{x}</b>"



import base64
import dataclasses
import io
import logging
import random
import re
import string
from abc import abstractmethod, ABC
from enum import Enum
from pathlib import Path
from typing import Dict, Union, List, Optional, Tuple

import math
import plotly.graph_objects as go

from multiqc.plots.plotly import check_plotly_version
from multiqc.utils import mqc_colour, config, report

logger = logging.getLogger(__name__)


class PlotType(Enum):
    """
    Plot labels used in custom content, as well as in JS to load plot into Plotly-js
    """

    LINE = "xy_line"
    VIOLIN = "violin"
    BAR = "bar_graph"
    SCATTER = "scatter"
    HEATMAP = "heatmap"
    BOX = "box"


@dataclasses.dataclass
class BaseDataset(ABC):
    """
    Structured version of dataset config dictionary
    """

    plot: "Plot"
    label: str
    uid: str
    dconfig: Dict  # user dataset-specific configuration
    layout: Dict  # update when a datasets toggle is clicked, or percentage switch is unselected
    trace_params: Dict
    pct_range: Dict

    def dump_for_javascript(self) -> Dict:
        d = {k: v for k, v in self.__dict__.items()}
        del d["plot"]
        return d

    def create_figure(self, layout: go.Layout, is_log=False, is_pct=False):
        """
        To be overridden by specific plots: create a Plotly figure for a dataset, update layout if needed.
        """
        raise NotImplementedError


class Plot(ABC):
    """Structured version of plot config dictionary"""

    # Default width for flat plots
    FLAT_PLOT_WIDTH = 1100

    def __init__(
        self,
        plot_type: PlotType,
        pconfig: Dict,
        n_datasets: int,
        id: Optional[str] = None,
    ):
        check_plotly_version()

        if n_datasets == 0:
            raise ValueError("No datasets to plot")

        self.id = id or pconfig.get("id")
        if self.id is None:  # ID of the plot group
            uniq_suffix = "".join(random.sample(string.ascii_lowercase, 10))
            is_static_suf = "static_" if config.plots_force_flat else ""
            self.id = f"mqc_{is_static_suf}plot_{uniq_suffix}"

        self.plot_type = plot_type
        self.flat = config.plots_force_flat or (
            not config.plots_force_interactive and n_datasets > config.plots_flat_numseries
        )
        self.pconfig = pconfig

        # Counts / Percentages / Log10 switch
        self.add_log_tab = pconfig.get("logswitch", False) and plot_type in [PlotType.BAR, PlotType.LINE]
        self.add_pct_tab = pconfig.get("cpswitch", True) is not False and plot_type == PlotType.BAR
        self.l_active = self.add_log_tab and pconfig.get("logswitch_active") is True
        self.p_active = self.add_pct_tab and pconfig.get("cpswitch_c_active", True) is not True

        # Per-dataset configurations
        self.datasets: List[BaseDataset] = [
            BaseDataset(
                self,
                label=str(i + 1),
                uid=self.id,
                dconfig=dict(),
                layout=dict(),
                trace_params=dict(),
                pct_range=dict(  # range for the percentage view for each axis
                    xaxis=dict(min=0, max=100),
                    yaxis=dict(min=0, max=100),
                ),
            )
            for i in range(n_datasets)
        ]

        height = self.pconfig.get("height", 500)
        width = self.pconfig.get("width")
        if self.pconfig.get("square"):
            width = height

        title = self.pconfig.get("table_title", self.pconfig.get("title"))
        if not title:
            if config.strict:
                errmsg = f"LINT: 'title' is missing from plot config for plot '{self.id}'"
                logger.error(errmsg)
                report.lint_errors.append(errmsg)

        self.layout = go.Layout(
            title=go.layout.Title(
                text=title,
                xanchor="center",
                x=0.5,
                font=dict(size=20),
            ),
            xaxis=go.layout.XAxis(
                gridcolor="rgba(0,0,0,0.05)",
                zerolinecolor="rgba(0,0,0,0.05)",
                color="rgba(0,0,0,0.3)",  # axis labels
                tickfont=dict(size=10, color="rgba(0,0,0,1)"),
                automargin=True,  # auto-expand axis to fit the tick labels
            ),
            yaxis=go.layout.YAxis(
                gridcolor="rgba(0,0,0,0.05)",
                zerolinecolor="rgba(0,0,0,0.05)",
                color="rgba(0,0,0,0.3)",  # axis labels
                tickfont=dict(size=10, color="rgba(0,0,0,1)"),
                automargin=True,  # auto-expand axis to fit the tick labels
            ),
            height=height,
            width=width,
            paper_bgcolor="rgba(0,0,0,0)",
            plot_bgcolor="rgba(0,0,0,0)",
            font=dict(family="'Lucida Grande', 'Open Sans', verdana, arial, sans-serif"),
            colorway=mqc_colour.mqc_colour_scale.COLORBREWER_SCALES["plot_defaults"],
            autosize=True,
            margin=go.layout.Margin(
                pad=5,  # pad sample names in a bar graph a bit
                t=50,  # more compact title
                r=15,  # remove excessive whitespace on the right
                b=65,  # remove excessive whitespace on the bottom
                l=60,  # remove excessive whitespace on the left
            ),
            hoverlabel=go.layout.Hoverlabel(
                namelength=-1,  # do not crop sample names inside hover label <extra></extra>
            ),
            modebar=go.layout.Modebar(
                bgcolor="rgba(0, 0, 0, 0)",
                color="rgba(0, 0, 0, 0.5)",
                activecolor="rgba(0, 0, 0, 1)",
            ),
            showlegend=self.flat,
            legend=go.layout.Legend(
                orientation="h",
                yanchor="top",
                y=-0.15,
                xanchor="center",
                x=0.5,
            )
            if self.flat
            else None,
        )
        # Layout update for the counts/percentage switch
        self.pct_axis_update = dict(
            ticksuffix="%",
            hoverformat=".1f",
        )

        self._axis_controlled_by_switches = self.axis_controlled_by_switches()
        if self.pconfig.get("xLog"):
            self.layout.xaxis.type = "log"
            if "xaxis" in self._axis_controlled_by_switches:
                self._axis_controlled_by_switches.remove("xaxis")
        if self.pconfig.get("yLog"):
            self.layout.yaxis.type = "log"
            if "yaxis" in self._axis_controlled_by_switches:
                self._axis_controlled_by_switches.remove("yaxis")
        if self.add_log_tab and self.l_active:
            for axis in self._axis_controlled_by_switches:
                self.layout[axis].type = "log"

        dconfigs: List[Union[str, Dict[str, str]]] = pconfig.get("data_labels", [])
        for idx, dataset in enumerate(self.datasets):
            if n_datasets > 1:
                dataset.uid += f"_{idx + 1}"

            if idx < len(dconfigs):
                dconfig = dconfigs[idx]
            else:
                dconfig = {}

            if not isinstance(dconfig, (str, dict)):
                logger.warning(f"Invalid data_labels type: {type(dconfig)}. Must be a string or a dict.")
            dconfig = dconfig if isinstance(dconfig, dict) else {"name": dconfig}
            dataset.label = dconfig.get("name", dconfig.get("label", idx + 1))
            if "ylab" not in dconfig and "ylab" not in self.pconfig:
                dconfig["ylab"] = dconfig.get("name", dconfig.get("label"))

            dataset.layout, dataset.trace_params = _dataset_layout(pconfig, dconfig, self.tt_label())
            dataset.dconfig = dconfig

    @staticmethod
    def axis_controlled_by_switches() -> List[str]:
        """
        Return a list of axis names that are controlled by the log10 scale and percentage
        switch buttons, e.g. ["yaxis"]
        """
        return []

    @staticmethod
    def tt_label() -> Optional[str]:
        """Default tooltip label"""
        return None

    def __repr__(self):
        d = {k: v for k, v in self.__dict__.items() if k not in ("datasets", "layout")}
        return f"<{self.__class__.__name__} {self.id} {d}>"

    def add_to_report(self, report) -> str:
        """
        Build and add the plot data to the report, return an HTML wrapper.
        """
        # Setting IDs again now that we have "report" object to guarantee uniqueness
        self.id = report.save_htmlid(self.id)
        for ds in self.datasets:
            ds.uid = self.id
            if len(self.datasets) > 1:  # for flat plots, each dataset will have its own unique ID
                ds.uid = report.save_htmlid(f"{self.id}_{ds.label}", skiplint=True)

        if self.flat:
            html = self.flat_plot()
        else:
            html = self.interactive_plot(report)
        return html

    def interactive_plot(self, report) -> str:
        html = '<div class="mqc_hcplot_plotgroup">'

        html += self.__control_panel()

        # This width only affects the space before plot is rendered, and the initial
        # height for the resizing function. For the actual plot container, Plotly will
        # re-calculate the wrapper size after rendering.
        height = self.layout.height
        height_style = f'style="height:{height + 7}px"' if height else ""
        html += f"""
        <div class="hc-plot-wrapper hc-{self.plot_type.value}-wrapper" id="{self.id}-wrapper" {height_style}>
            <div id="{self.id}" class="hc-plot hc-{self.plot_type.value}-plot not_rendered"></div>
            <div class="created-with-multiqc">Created with MultiQC</div>
        </div>"""

        html += "</div>"

        # Saving compressed data for JavaScript to pick up and uncompress.
        dump = self.dump_for_javascript()
        report.plot_data[self.id] = dump
        return html

    def flat_plot(self) -> str:
        html = "".join(
            [
                '<p class="text-info">',
                "<small>" '<span class="glyphicon glyphicon-picture" aria-hidden="true"></span> ',
                "Flat image plot. Toolbox functions such as highlighting / hiding samples will not work ",
                '(see the <a href="http://multiqc.info/docs/#flat--interactive-plots" target="_blank">docs</a>).',
                "</small>",
                "</p>",
            ]
        )
        html += f'<div class="mqc_mplplot_plotgroup" id="plotgroup-{self.id}" data-pid={self.id}>'

        if not config.simple_output:
            html += self.__control_panel()

        # Go through datasets creating plots
        for ds_idx, dataset in enumerate(self.datasets):
            if self.pconfig.get("save_data_file", True) and self.id != "general_stats_table":
                self.save_data_file(dataset)

            html += self._fig_to_static_html(
                self._make_flat_fig(dataset),
                active=ds_idx == 0 and not self.p_active and not self.l_active,
                uid=dataset.uid if not self.add_log_tab and not self.add_pct_tab else f"{dataset.uid}-cnt",
            )
            if self.add_pct_tab:
                html += self._fig_to_static_html(
                    self._make_flat_fig(dataset, is_pct=True),
                    active=ds_idx == 0 and self.p_active,
                    uid=f"{dataset.uid}-pct",
                )
            if self.add_log_tab:
                html += self._fig_to_static_html(
                    self._make_flat_fig(dataset, is_log=True),
                    active=ds_idx == 0 and self.l_active,
                    uid=f"{dataset.uid}-log",
                )
            if self.add_pct_tab and self.add_log_tab:
                html += self._fig_to_static_html(
                    self._make_flat_fig(dataset, is_pct=True, is_log=True),
                    active=ds_idx == 0 and self.p_active and self.l_active,
                    uid=f"{dataset.uid}-pct-log",
                )

        html += "</div>"
        return html

    def _btn(self, cls: str, label: str, data_attrs: Dict[str, str] = None, pressed: bool = False) -> str:
        """Build a switch button for the plot."""
        data_attrs = data_attrs.copy() if data_attrs else {}
        if "pid" not in data_attrs:
            data_attrs["pid"] = self.id
        data_attrs = " ".join([f'data-{k}="{v}"' for k, v in data_attrs.items()])
        return f'<button class="btn btn-default btn-sm {cls} {"active" if pressed else ""}" {data_attrs}>{label}</button>\n'

    def buttons(self) -> List[str]:
        """
        Build buttons for control panel
        """
        switch_buttons = ""
        cls = "mpl_switch_group" if self.flat else "interactive-switch-group"
        # Counts / percentages / log10 switches
        if self.add_pct_tab or self.add_log_tab:
            if self.add_pct_tab:
                switch_buttons += self._btn(
                    cls=f"{cls} percent-switch",
                    label=self.pconfig.get("cpswitch_percent_label", "Percentages"),
                    pressed=self.p_active,
                )
            if self.add_log_tab:
                switch_buttons += self._btn(
                    cls=f"{cls} log10-switch",
                    label=self.pconfig.get("logswitch_label", "Log10"),
                    pressed=self.l_active,
                )

        # Buttons to cycle through different datasets
        if len(self.datasets) > 1:
            switch_buttons += f'<div class="btn-group {cls} dataset-switch-group">\n'
            for ds_idx, ds in enumerate(self.datasets):
                data_attrs = {
                    "dataset-index": ds_idx,
                    # For flat plots, we will generate separate flat images for each
                    # dataset and view, so have to save individual image IDs.
                    "dataset-uid": ds.uid,
                }
                switch_buttons += self._btn(
                    cls="mr-auto",
                    label=ds.label,
                    data_attrs=data_attrs,
                    pressed=ds_idx == 0,
                )
            switch_buttons += "</div>\n\n"

        export_btn = ""
        if not self.flat:
            export_btn = self._btn(cls="export-plot", label="Export Plot")
        return [switch_buttons, export_btn]

    def __control_panel(self) -> str:
        """
        Add buttons: percentage on/off, log scale on/off, datasets switch panel
        """
        buttons = "\n".join(self.buttons())
        html = f"<div class='row'>\n<div class='col-xs-12'>\n{buttons}\n</div>\n</div>\n\n"
        return html

    def dump_for_javascript(self) -> Dict:
        """Serialise the plot data to pick up in plotly-js"""
        return {
            "id": self.id,
            "layout": self.layout.to_plotly_json(),
            "datasets": [d.dump_for_javascript() for d in self.datasets],
            "plot_type": self.plot_type.value,
            "pct_axis_update": self.pct_axis_update,
            "axis_controlled_by_switches": self._axis_controlled_by_switches,
            "p_active": self.p_active,
            "l_active": self.l_active,
            "square": self.pconfig.get("square"),
            "config": self.pconfig,  # for megaqc
        }

    @abstractmethod
    def save_data_file(self, data: BaseDataset) -> None:
        """
        Save dataset to disk.
        """

    def _make_flat_fig(self, dataset: BaseDataset, is_log=False, is_pct=False) -> go.Figure:
        """
        Create a Plotly Figure object.
        """
        layout = go.Layout(self.layout.to_plotly_json())  # make a copy
        layout.update(**dataset.layout)
        layout.width = layout.width or Plot.FLAT_PLOT_WIDTH
        for axis in self.axis_controlled_by_switches():
            layout[axis].type = "linear"
            minval = layout[axis].autorangeoptions["minallowed"]
            maxval = layout[axis].autorangeoptions["maxallowed"]
            if is_pct:
                layout[axis].update(self.pct_axis_update)
                minval = dataset.pct_range.get(axis, {}).get("min", 0)
                maxval = dataset.pct_range.get(axis, {}).get("max", 100)
            if is_log:
                layout[axis].type = "log"
                minval = math.log10(minval) if minval is not None and minval > 0 else None
                maxval = math.log10(maxval) if maxval is not None and maxval > 0 else None
            layout[axis].autorangeoptions["minallowed"] = minval
            layout[axis].autorangeoptions["maxallowed"] = maxval
        return dataset.create_figure(layout, is_log, is_pct)

    @staticmethod
    def _fig_to_static_html(
        fig: go.Figure,
        active: bool,
        uid: str,
    ) -> str:
        """
        Build one static image, return an HTML wrapper.
        """
        assert fig.layout.width
        write_kwargs = dict(
            width=fig.layout.width,  # While interactive plots take full width of screen,
            # for the flat plots we explicitly set width
            height=fig.layout.height,
            scale=2,  # higher detail (retina display)
        )

        # Save the plot to the data directory if export is requested
        if config.export_plots:
            for file_ext in config.export_plot_formats:
                plot_fn = Path(config.plots_dir) / file_ext / f"{uid}.{file_ext}"
                plot_fn.parent.mkdir(parents=True, exist_ok=True)
                if file_ext == "svg":
                    # Cannot add logo to SVGs
                    fig.write_image(plot_fn, **write_kwargs)
                else:
                    img_buffer = io.BytesIO()
                    fig.write_image(img_buffer, **write_kwargs)
                    img_buffer = Plot.add_logo(img_buffer, format=file_ext)
                    with open(plot_fn, "wb") as f:
                        f.write(img_buffer.getvalue())
                    img_buffer.close()

        # Now writing the PNGs for the HTML
        if config.development:
            # Using file written in the config.export_plots block above
            img_src = Path(config.plots_dir_name) / "png" / f"{uid}.png"
        else:
            img_buffer = io.BytesIO()
            fig.write_image(img_buffer, **write_kwargs)
            img_buffer = Plot.add_logo(img_buffer, format="PNG")
            # Convert to a base64 encoded string
            b64_img = base64.b64encode(img_buffer.getvalue()).decode("utf8")
            img_src = f"data:image/png;base64,{b64_img}"
            img_buffer.close()

        # Should this plot be hidden on report load?
        hiding = "" if active else ' style="display:none;"'
        return "".join(
            [
                f'<div class="mqc_mplplot" id="{uid}"{hiding}>',
                f'<img src="{img_src}" height="{fig.layout.height}px" width="{fig.layout.width}px"/>',
                "</div>",
            ]
        )

    @staticmethod
    def add_logo(
        img_buffer: io.BytesIO,
        format: str = "png",
        text: str = "Created with MultiQC",
        font_size: int = 16,
    ) -> io.BytesIO:
        try:
            from PIL import Image, ImageDraw

            # Load the image from the BytesIO object
            image = Image.open(img_buffer)

            # Create a drawing context
            draw = ImageDraw.Draw(image)

            # Define the text position. In order to do that, first calculate the expected
            # text block width, given the font size.
            # noinspection PyArgumentList
            text_width: float = draw.textlength(text, font_size=font_size)
            position: Tuple[int, int] = (image.width - int(text_width) - 3, image.height - 30)

            # Draw the text
            draw.text(position, text, fill="#9f9f9f", font_size=font_size)

            # Save the image to a BytesIO object
            output_buffer = io.BytesIO()
            image.save(output_buffer, format=format)
            output_buffer.seek(0)

        except Exception as e:
            logger.warning(f"Failure adding logo to the plot: {e}")
            output_buffer = img_buffer

        return output_buffer

    @staticmethod
    def _get_logo() -> str:
        """
        Return the MultiQC logo as a base64 encoded string.
        """
        # logo_path = Path(__file__).parent.parent / "assets" / "img" / "MultiQC_logo.png"
        logo_path = Path(__file__).parent.parent / "assets" / "img" / "MultiQC_logo_dark.svg"
        with open(logo_path, "rb") as f:
            data = f.read()
        return "data:image/svg+xml;base64," + base64.b64encode(data).decode()


def rename_deprecated_highcharts_keys(conf: Dict) -> Dict:
    """
    Rename the deprecated HighCharts-specific terminology in a config.
    """
    conf = conf.copy()
    if "yCeiling" in conf:
        conf["yaxis"] = conf.pop("y_ceiling")
    if "xAxis" in conf:
        conf["xaxis"] = conf.pop("xAxis")
    if "tooltip" in conf:
        conf["hovertemplate"] = conf.pop("tooltip")
    return conf


def _dataset_layout(
    pconfig: Dict,
    dconfig: Dict,
    default_tt_label: Optional[str],
) -> Tuple[Dict, Dict]:
    """
    Given plot config and dataset config, set layout and trace params.
    """
    pconfig = pconfig.copy()
    pconfig.update(dconfig)

    # Format on-hover tooltips
    ysuffix = pconfig.get("ysuffix", pconfig.get("tt_suffix"))
    ylabformat = pconfig.get("ylab_format", pconfig.get("yLabFormat"))
    if ylabformat and "%" in ylabformat:
        ysuffix = "%"

    xsuffix = pconfig.get("xsuffix")
    xlabformat = pconfig.get("xlab_format", pconfig.get("xLabFormat"))
    if xlabformat and "%" in xlabformat:
        xsuffix = "%"

    if "tt_label" in pconfig:
        # clean label, add missing <br> into the beginning, and populate tt_suffix if missing
        tt_label = pconfig["tt_label"]
        tt_label = _clean_config_tt_label(tt_label)

        # if "%" character is inside the Y label, add suffix to the Y ticks as well
        parts = tt_label.split("%{")
        for part in parts:
            if ysuffix is None and part.startswith("y") and "}" in part:
                ysuffix = part.split("}")[1].replace("</b>", "").replace(":", "").rstrip()
                if "," in ysuffix:
                    ysuffix = ysuffix.split(",")[0]
            elif xsuffix is None and part.startswith("x") and "}" in part:
                xsuffix = part.split("}")[1].replace("</b>", "").replace(":", "").rstrip()
                if "," in xsuffix:
                    xsuffix = xsuffix.split(",")[0]

        # As the suffix will be added automatically for the simple format ({y}), remove it from the label
        if ysuffix is not None and "{y}" + ysuffix in tt_label:
            tt_label = tt_label.replace("{y}" + ysuffix, "{y}")
        if xsuffix is not None and "{x}" + xsuffix in tt_label:
            tt_label = tt_label.replace("{x}" + xsuffix, "{x}")

        # add missing line break between the sample name and the key-value pair
        if not tt_label.startswith("<br>"):
            tt_label = "<br>" + tt_label
    else:
        tt_label = default_tt_label

    if tt_label:
        hovertemplate = "<b>%{text}</b>" + tt_label + "<extra></extra>"
    else:
        hovertemplate = None

    # `hoverformat` describes how plain "{y}" or "{x}" are formatted in `hovertemplate`
    tt_decimals = pconfig.get("tt_decimals", pconfig.get("decimalPlaces"))
    y_hoverformat = f",.{tt_decimals}f" if tt_decimals is not None else None

    layout = dict(
        title=dict(text=pconfig.get("title")),
        xaxis=dict(
            hoverformat=None,
            ticksuffix=xsuffix or "",
            title=dict(text=pconfig.get("xlab")),
            rangemode="tozero" if pconfig.get("xmin") == 0 else "normal",
            autorangeoptions=dict(
                clipmin=pconfig.get("xFloor"),
                clipmax=pconfig.get("xCeiling"),
                minallowed=pconfig.get("xmin"),
                maxallowed=pconfig.get("xmax"),
            ),
        ),
        yaxis=dict(
            hoverformat=y_hoverformat,
            ticksuffix=ysuffix or "",
            title=dict(text=pconfig.get("ylab")),
            rangemode="tozero" if pconfig.get("ymin") == 0 == 0 else "normal",
            autorangeoptions=dict(
                clipmin=pconfig.get("yFloor"),
                clipmax=pconfig.get("yCeiling"),
                minallowed=pconfig.get("ymin"),
                maxallowed=pconfig.get("ymax"),
            ),
        ),
    )

    trace_params = {}
    if hovertemplate:
        trace_params["hovertemplate"] = hovertemplate

    return layout, trace_params


def _clean_config_tt_label(tt_label: str) -> str:
    replace_d = {  # Convert HighCharts format to Plotly format
        "{point.x": "%{x",
        "{point.y": "%{y",
        "x:>": "x:",
        "y:>": "y:",
        "{point.category}": "%{x}",
        "<strong>": "<b>",
        "</strong>": "</b>",
        "<br/>": "<br>",
    }
    for k, v in replace_d.items():
        tt_label = tt_label.replace(k, v)
    return tt_label


def split_long_string(s: str, max_width=80) -> List[str]:
    """
    Split string into lines of max_width characters
    """
    lines = []
    current_line = ""
    words = re.split(r"(\W+)", s)
    for word in words:
        if len(current_line + word) <= max_width:
            current_line += word
        else:
            if current_line:
                lines.append(current_line)
            current_line = word

    if current_line:
        lines.append(current_line)

    return lines



from itertools import chain

from multiqc.plots import linegraph


def plot_bqhist(samples, file_type, **plot_args):
    """Create line graph plot of histogram data for BBMap 'bqhist' output.

    The 'samples' parameter could be from the bbmap mod_data dictionary:
    samples = bbmap.MultiqcModule.mod_data[file_type]
    """

    all_x = set()
    for item in sorted(chain(*[samples[sample]["data"].items() for sample in samples])):
        all_x.add(item[0])

    columns_to_plot = {
        #'Counts': {
        #    0: 'Read 1',
        #    9: 'Read 2',
        # },
        "Read 1 averages": {
            # 1: 'Min',
            # 2: 'Max',
            3: "Mean",
            # 4: '1st quartile',
            # 5: 'Median',
            # 6: '3rd quartile',
            # 7: 'Left (lower) whisker',
            # 8: 'Right (upper) whisker',
        },
        "Read 2 averages": {
            # 10: 'Min',
            # 11: 'Max',
            12: "Mean",
            # 13: '1st quartile',
            # 14: 'Median',
            # 15: '3rd quartile',
            # 16: 'Left (lower) whisker',
            # 17: 'Right (upper) whisker',
        },
    }

    plot_data = []
    for column_type in columns_to_plot:
        plot_data.append(
            {
                sample + "." + column_name: {
                    x: samples[sample]["data"][x][column] if x in samples[sample]["data"] else 0 for x in all_x
                }
                for sample in samples
                for column, column_name in columns_to_plot[column_type].items()
            }
        )

    plot_params = {
        "id": "bbmap-" + file_type + "_plot",
        "title": "BBTools: " + plot_args["plot_title"],
        "xlab": "Read position",
        "ylab": "Average quality score",
        "data_labels": [
            # {'name': 'Count histogram', 'ylab': 'Read count'},
            {
                "name": "Read 1",
            },
            {
                "name": "Read 2",
            },
        ],
    }

    plot_params.update(plot_args["plot_params"])
    plot = linegraph.plot(plot_data, plot_params)

    return plot
